#ifndef APP_H
#define APP_H

#include "general.h"
#include "memory.h"
#include "camera.h"
#include "event.h"
#include "kernel.h"
#include "input.h"
#include "sdlinput.h"
#include "glrender.h"
#include "texture.h"
#include "model.h"
#include "glsl.h"
#include "font.h"
#include "light.h"
#include "script.h"
#include "glwidget.h"
#include "terrain.h"

#include <time.h>

/*!
 * \file   app.h
 * \author  <baldurk@baldurk.org>
 * \date   Tue Jan 10 21:57:03 2006
 * 
 * \brief  A sample application using the engine  
 */

//! A class to use the engine
class Application : public EventCatcher
{
  public:
    //! A constructor to intialise variables
    Application() : framecount(0), timecount(0) {}
    //! Initialise the application
    bool Init();
    //! Inherited function called when events are received
    //! \param event the event received
    bool GotEvent(Event event);

  private:
    //! A function to update the application state
    //! \param milliseconds the number of milliseconds since the last update call
    bool Update(int milliseconds);
    //! A function to render the application
    bool Render(bool real = true);
    //! Quits the application
    bool Quit();

    void SetupCam(int t, int v, int i);

    Model *room;
    FPSCamera cam;
    Light light;

    FontRender font;

    Texture tex;

    KeyboardEvent keys;

    GLSL prog;

    int tri;
    Vectorf norm, right, up;

    struct Obj
    {
	Obj(Vectorf a=Vectorf(), Vectorf b=Vectorf(), float c=1.0f) :
	    pos(a), vel(b), mass(c) {}
	Vectorf pos;Vectorf vel;float mass;
    };vector<Obj> objects;

    //! the number of frames in the current 5 second period
    int framecount,
	// the number of milliseconds passed so far (up to 5 seconds)
	timecount;

    //! a counter (the time passed in milliseconds)
    float x,
	//! a counter (the time passed in milliseconds)
	y,
	//! a counter (the time passed in milliseconds)
	z;

    //! a pointer to the kernel
    Pointer<Kernel> kernel;
};

#endif
