#include "model.h"

/*!
 * \file   bkmodel.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Thu May 18 01:22:08 2006
 *
 * \brief  Specific implementation of a model for BKM files
 *  
 */

int BKMModel::Load(string Filename)
{
    FUNC("BKMModel::Load");
    FILE *fp = fopen(Filename.c_str(), "r");

    if(!fp)
    {
	Log << "'" << Filename << "': Invalid model filename" << endl;;
	ENDRETURN(0);
    }

    string line;

    fclose(fp);
    ENDRETURN(1);
}

/*
    //! the bones in the model
    vector<Bone*> Bones;

    //! the vertices in the model
    vector<Vertex> OrigVertices;

    //! the triangles in the model
    vector<Triangle> Triangles;

    //! the materials in the model
    vector<Material> Materials;
 */

int BKMModel::Write(string f)
{
    FUNC("BKMModel::Write");

    Log << "writing" << endl;

    int NumBones = Bones.size(),
	NumVerts = OrigVertices.size(),
	NumTris = Triangles.size(),
	NumMats = Materials.size();

    unsigned char *origbuf = new unsigned char[NumBones*sizeof(Bone) +
					       NumVerts*sizeof(Vertex) +
					       NumTris*sizeof(Triangle) +
					       NumMats*sizeof(Material)];

    for(int i=0; i < NumBones; i++)
	memcpy(origbuf+sizeof(Bone)*i, &(*Bones[i]), sizeof(Bone));
    
    int offset = NumBones*sizeof(Bone);

    memcpy(origbuf+offset, &OrigVertices, NumVerts*sizeof(Vertex));
    offset+=NumVerts*sizeof(Vertex);
    memcpy(origbuf+offset, &Triangles, NumTris*sizeof(Triangle));
    offset+=NumTris*sizeof(Triangle);
    memcpy(origbuf+offset, &Materials, NumMats*sizeof(Material));
    assert(offset+ NumMats*sizeof(Material) == 
	   NumBones*sizeof(Bone) +
	   NumVerts*sizeof(Vertex) +
	   NumTris*sizeof(Triangle) +
	   NumMats*sizeof(Material));

    unsigned long dstsize, srcsize = NumBones*sizeof(Bone) +
	NumVerts*sizeof(Vertex) +
	NumTris*sizeof(Triangle) +
	NumMats*sizeof(Material);

    Deflate(NULL, &dstsize, origbuf, srcsize);

    unsigned char *dstbuf = new unsigned char[dstsize];

    Deflate(dstbuf, &dstsize, origbuf, srcsize);

    delete[] origbuf;
    delete[] dstbuf;

    ENDRETURN(1);
}
