#include "camera.h"

/*!
 * \file   camera.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Tue Feb 14 16:51:22 2006
 * 
 * \brief  the implementation of camera classes
 */

bool FPSCamera::Init(float tspeed, float rspeed)
{
    FUNC("FPSCamera::Init");
    trans = Vectorf();rot = Vectorf();
    up = right = forward = Vectorf();
    up.y = right.x = 1.0f;forward.z = -1.0f;
    Reinit(tspeed, rspeed);
    ENDRETURN(true);
}

bool FPSCamera::Reinit(float tspeed, float rspeed)
{
    FUNC("FPSCamera::Reinit");
    TSpeed = tspeed;RSpeed = rspeed;
    ENDRETURN(true);
}

bool FPSCamera::Update(float time)
{
    FUNC("FPSCamera::Update");
    Time = time;
    ENDRETURN(true);
}

void FPSCamera::Move(bool forward, bool backward,
		     bool up, bool down,
		     bool left, bool right)
{
    FUNC("FPSCamera::Keys");

    Vectorf dt;

    if(up)
	dt.y += 1.0f;
    if(down)
	dt.y -= 1.0f;
    if(forward)
    {
	dt.x += sinf(DEGTORAD(rot.y)) * cosf(DEGTORAD(rot.x));
	dt.y -= sinf(DEGTORAD(rot.x));
	dt.z -= cosf(DEGTORAD(rot.y)) * cosf(DEGTORAD(rot.x));
    }
    if(backward)
    {
	dt.x -= sinf(DEGTORAD(rot.y)) * cosf(DEGTORAD(rot.x));
	dt.y += sinf(DEGTORAD(rot.x));
	dt.z += cosf(DEGTORAD(rot.y)) * cosf(DEGTORAD(rot.x));
    }
    if(left)
    {
	dt.x -= cosf(DEGTORAD(rot.y));
	dt.z -= sinf(DEGTORAD(rot.y));
    }
    if(right)
    {
	dt.x += cosf(DEGTORAD(rot.y));
	dt.z += sinf(DEGTORAD(rot.y));
    }

    if(dt.length() > 1.0f)
	dt.normalise();

    trans += dt * Time * TSpeed;

    END();
}

void FPSCamera::Look(float dx, float dy, float dz)
{
    FUNC("FPSCamera::Mouse");
    
    if(rot.x - RSpeed * dy > -89.9f &&
       rot.x - RSpeed * dy < 89.9f)
	rot.x -= RSpeed * dy;
    else
	dy = 0.0f;
    rot.y += RSpeed * dx;
    rot.z += RSpeed * dz;

    END();
}

bool FPSCamera::Translate(Vectorf Trans)
{
    FUNC("FPSCamera::Translate");
    trans = Trans;
    ENDRETURN(true);
}

bool FPSCamera::Transform()
{
    FUNC("FPSCamera::Transform");
    glLoadIdentity();
    
    glRotatef(rot.x, 1.0f, 0.0f, 0.0f);
    glRotatef(rot.y, 0.0f, 1.0f, 0.0f);
    glRotatef(rot.z, 0.0f, 0.0f, 1.0f);
    glTranslatef(-trans.x, -trans.y, -trans.z);
    
    /*
    gluLookAt(trans.x, trans.y, trans.z,
	      trans.x+forward.x, trans.y+forward.y, trans.z+forward.z,
	      up.x, up.y, up.z);
    */
    ENDRETURN(true);
}

bool FPSCamera::LookAt(Vectorf pos, Vectorf look, Vectorf Up)
{
    FUNC("FPSCamera::LookAt");
    trans = pos;forward = look-pos;up = Up;

    trans = pos;
    Vectorf eye = look-pos;
    eye.normalise();

    Vectorf eyexz = Vectorf(eye.x, 0.0f, eye.z);
    eyexz.normalise();

    Vectorf eyeyz = Vectorf(0.0f, eye.y, eye.z);
    eyeyz.normalise();

    if(fabs(eye.x) < 0.001f && fabs(eye.z) < 0.001f)
	rot.y = 0.0f;
    else if(fabs(eye.x) > fabs(eye.z) && eye.x < 0.0f)
	rot.y = -90.0f+RADTODEG(atanf(fabs(eye.z)/fabs(eye.x)));
    else if(fabs(eye.x) > fabs(eye.z) && eye.x > 0.0f)
	rot.y = 90.0f-RADTODEG(atanf(fabs(eye.z)/fabs(eye.x)));
    else if(eye.z > 0.0f)
	rot.y = 180.0f+RADTODEG(atanf(fabs(eye.x)/fabs(eye.z)));
    else if(eye.z < 0.0f)
	rot.y = -RADTODEG(atanf(fabs(eye.x)/fabs(eye.z)));

    if(fabs(eye.y) < 0.001f && fabs(eye.z) < 0.001f)
	rot.x = 0.0f;
    else if(fabs(eye.y) > fabs(eye.z) && eye.y < 0.0f)
	rot.x = 90.0f-RADTODEG(atanf(fabs(eye.z)/fabs(eye.y)));
    else if(fabs(eye.y) > fabs(eye.z) && eye.y > 0.0f)
	rot.x = +90.0f+RADTODEG(atanf(fabs(eye.z)/fabs(eye.y)));
    else if(eye.z > 0.0f)
	rot.x = RADTODEG(atanf(fabs(eye.y)/fabs(eye.z)));
    else if(eye.z < 0.0f)
	rot.x = -RADTODEG(atanf(fabs(eye.y)/fabs(eye.z)));

    ENDRETURN(true);
}

