#include "script.h"

/*!
 * \file   compile.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:40:46 2006
 * 
 * \brief  functions to compile text to bytecode  
 */

bool Script::Start(string str, string test)
{
    return (str.find(test) == 0);
}

bool Script::Contains(string str, string test)
{
    return (str.find(test) != string::npos);
}

#define TrimFirst(string) string = string.substr(1)
#define TrimLast(string) string = string.substr(0,string.length()-1)

bool Script::StartNumber(string data)
{
    return Start(data, "0") || Start(data, "1") || Start(data, "2") ||
	Start(data, "3") || Start(data, "4") || Start(data, "5") || 
	Start(data, "6") || Start(data, "7") || Start(data, "8") || 
	Start(data, "9") || Start(data, ".") || Start(data, "-");
}

string Script::GetNumber(string &data)
{
    FUNC("Script::GetNumber");
    string ret = "";
    while(StartNumber(data))
    {
	ret += data[0];
	TrimFirst(data);
    }
    ENDRETURN(ret);
}

bool Script::StartSpace(string data)
{
    return Start(data, " ") || Start(data, "\t") || Start(data, "\n");
}

bool Script::IsSpace(string data)
{
    FUNC("Script::IsSpace");
    for(size_t i=0; i < data.length(); i++)
    {
	if(!StartSpace(data.substr(i)))
	    ENDRETURN(false);
    }
    ENDRETURN(true);
}

string Script::GetSpace(string &data)
{
    FUNC("Script::GetSpace");
    string ret = "";
    while(StartSpace(data))
    {
	ret += data[0];
	TrimFirst(data);
    }
    ENDRETURN(ret);
}

bool Script::StartVariable(string data)
{
    return isalpha(data[0]) || Start(data, "-") || Start(data, "_") || Start(data, ":");
}

string Script::GetVariableName(string &data)
{
    FUNC("Script::GetVariableName");
    string ret = "";
    while(StartVariable(data))
    {
	ret += data[0];
	TrimFirst(data);
    }
    ENDRETURN(ret);
}

string Script::GetVariable(string &data)
{
    FUNC("Script::GetVariable");
    string ret = "";
    while(StartVariable(data) || Start(data, "["))
    {
	while(data[0] == '[')
	{
	    ret += data[0];
	    TrimFirst(data);
	    ret += GetNumber(data);
	}
	ret += data[0];
	TrimFirst(data);
    }
    ENDRETURN(ret);
}

bool Script::StartString(string data)
{
    return Start(data, "\"");
}

string Script::GetString(string &data)
{
    FUNC("Script::GetString");
    string ret = "";
    if(StartString(data))
    {
	TrimFirst(data);
	while(data[0] != '"')
	{
	    if(data[0] == '\'')
	    {
		ret += data[1];
		TrimFirst(data);
		TrimFirst(data);
	    }
	    else
	    {
		ret += data[0];
		TrimFirst(data);
	    }
	}
	TrimFirst(data);
    }

    ENDRETURN(ret);
}

// compile a string expression to a series of tokens.
// this just grabs string segments and converts to tokens naively.
// that means no checking of what the tokens mean
bool Script::Compile(vector<token> &ts, string expr)
{
    FUNC("Script::Compile");

#define ADDTOKEN(t) if(func)func->tokens.push_back(t);else ts.push_back(t);

    while(expr.size())
    {
	// trim whitespace
	if(Start(expr, " ") || Start(expr, "\n") || Start(expr, "\t"))
	{
	    TrimFirst(expr);
	    continue;
	}

	// get a number
	bool processed = false;
	if(StartNumber(expr))
	{
	    string num = GetNumber(expr);
	    istringstream iss;float f; iss.str(num); iss >> f;
	    ScriptVariable var;
	    if((int)f == f)
		var = (int)f;
	    else
		var = f;

	    ADDTOKEN(token(VALUE, var));

	    processed = true;
	}

	// get a string
	if(!processed && StartString(expr))
	{
	    string str = GetString(expr);
	    ADDTOKEN(token(STRING, str));

	    processed = true;
	}

	// get a return statement
	if(!processed && Start(expr, "return "))
	{
	    expr = expr.substr(7);
	    ADDTOKEN(RET);
	    processed = true;
	}

	// get a "special" operation string
	if(!processed &&
	   (Start(expr, "func ") || Start(expr, "while") ||
	    Start(expr, "for") || Start(expr, "if")))
	{
	    token t;
	    // get the type of operation
	    if(Start(expr, "while"))
	    {
		expr = expr.substr(5);t.type = WHILE;
	    }
	    if(Start(expr, "func "))
	    {
		expr = expr.substr(5);t.type = FUNCDEF;
		GetSpace(expr);
		string funcname = GetVariable(expr);
		t.name = funcname;
	    }
	    if(Start(expr, "for"))
	    {
		expr = expr.substr(3);t.type = FOR;
	    }
	    if(Start(expr, "if"))
	    {
		expr = expr.substr(2);t.type = IF;
	    }
	    GetSpace(expr);

	    // parse what's in the ()s
	    if(expr[0] == '(')
	    {
		int bracketlevel = 1;
		size_t i;
		for(i=expr.find("(")+1; i < expr.length(); i++)
		{
		    if(bracketlevel == 0)
			break;
		    
		    if(expr[i] == '(')
			bracketlevel++;
		    if(expr[i] == ')')
			bracketlevel--;
		}

		string brackexp = expr.substr(expr.find("("), i-expr.find("("));
		TrimFirst(brackexp);TrimLast(brackexp);

		// trim off the brackets and get the contents

		expr = expr.substr(i);

		// if it's a function, get a pointer to the function's location
		if(t.type == FUNCDEF)
		{
		    func = &scriptfuncs[t.name];
		    blocklevel = 0;
		}

		int numparams = 0, pos = 0;
		pos = brackexp.find(",", pos);

		while(pos != (signed)string::npos)
		{
		    numparams++;
		    pos = brackexp.find(",", pos+1);
		}

		if(numparams > 0)
		    numparams++;
		else if(numparams == 0 && !IsSpace(brackexp))
		    numparams++;

		// get the parameters
		if(numparams)
		{
		    if(brackexp.find(",") == string::npos)
		    {
			scrfunc *savefunc = func;
			func = NULL;
			Compile(t.parameters, brackexp);
			func = savefunc;
			t.parameters.push_back(NOP);
		    }
		    else
		    {
			scrfunc *savefunc = func;
			func = NULL;
			Compile(t.parameters, brackexp.substr(0, brackexp.find(",")));
			func = savefunc;
			t.parameters.push_back(NOP);
		    }

		    string work = brackexp.substr(brackexp.find(",")+1);
		    for(int i=1; i < numparams; i++)
		    {
			scrfunc *savefunc = func;
			func = NULL;
			Compile(t.parameters, work.substr(0, work.find(",")));
			func = savefunc;
			t.parameters.push_back(NOP);
			work = work.substr(work.find(",")+1);
		    }
		}
		if(t.type != FUNCDEF || !func)
		{
		    ADDTOKEN(t);
		    ADDTOKEN(EOL);
		}
		else
		{
		    func->parameters = t.parameters;
		}
	    }
	    processed = true;
	}

	// get a variable
	if(!processed && StartVariable(expr))
	{
	    string var = GetVariable(expr);
	    GetSpace(expr);
	    // if there's a (), it's a function call.
	    // get the parameters passed
	    if(expr[0] == '(')
	    {
		token t(FUNCCALL);
		t.name = var;

		int bracketlevel = 1;
		size_t i;
		for(i=expr.find("(")+1; i < expr.length(); i++)
		{
		    if(bracketlevel == 0)
			break;

		    if(expr[i] == '(')
			bracketlevel++;
		    if(expr[i] == ')')
			bracketlevel--;
		}

		// get the expression in the brackets
		string brackexp = expr.substr(expr.find("("), i-expr.find("("));
		TrimFirst(brackexp);TrimLast(brackexp);

		expr = expr.substr(i);

		int numparams = 0, pos = 0;
		pos = brackexp.find(",", pos);

		while(pos != (signed)string::npos)
		{
		    numparams++;
		    pos = brackexp.find(",", pos+1);
		}

		if(numparams > 0)
		    numparams++;
		else if(numparams == 0 && !IsSpace(brackexp))
		    numparams++;

		// parse through the expressions recursively
		if(numparams)
		{
		    if(brackexp.find(",") == string::npos)
		    {
			scrfunc *savefunc = func;
			func = NULL;
			Compile(t.parameters, brackexp);
			func = savefunc;
			t.parameters.push_back(NOP);
		    }
		    else
		    {
			scrfunc *savefunc = func;
			func = NULL;
			Compile(t.parameters, brackexp.substr(0, brackexp.find(",")));
			func = savefunc;
			t.parameters.push_back(NOP);
		    }

		    string work = brackexp.substr(brackexp.find(",")+1);
		    for(int i=1; i < numparams; i++)
		    {
			scrfunc *savefunc = func;
			func = NULL;
			Compile(t.parameters, work.substr(0, work.find(",")));
			func = savefunc;
			t.parameters.push_back(NOP);
			work = work.substr(work.find(",")+1);
		    }
		}
		else
		{
		    t.parameters.push_back(NOP);
		}

		ADDTOKEN(t);
	    }
	    else
	    {
		token t(VARNAME);
		t.name = var;
		ADDTOKEN(t);
	    }

	    processed = true;
	}

	// go through the ops, checking the translation table and translating
	// operations
	for(int i=0; i < LASTOPNUM; i++)
	{
	    if(!processed && Start(expr, toktrans[i]))
	    {
		processed = true;
		ADDTOKEN(token(toktype(i)));
		for(size_t l=0; l < strlen(toktrans[i]); l++)
		    TrimFirst(expr);
	    }
	}

	// get an end of line marker
	if(!processed && Start(expr, ";"))
	{
	    processed = true;
	    ADDTOKEN(EOL);
	    TrimFirst(expr);
	}

	// get a block open
	if(!processed && Start(expr, "{"))
	{
	    processed = true;
	    blocklevel++;
	    ADDTOKEN(token(BLOCKOPEN));
	    ADDTOKEN(EOL);
	    TrimFirst(expr);
	}

	// get a block close
	if(!processed && Start(expr, "}"))
	{
	    processed = true;
	    blocklevel--;
	    ADDTOKEN(token(BLOCKCLOSE));
	    ADDTOKEN(EOL);

	    if(blocklevel == 0)
		func = NULL;
	    TrimFirst(expr);
	}

	if(!processed)
	    TrimFirst(expr);
    }
    ENDRETURN(true);
}

// output a vector of tokens
ostream &operator <<(ostream &o, vector<token> toks)
{
    FUNC("Script::<<");
    for(size_t i=0; i < toks.size(); i++)
    {
	if(toks[i].type < NUMTOKS && toks[i].type >= 0)
	    o << toktrans[int(toks[i].type)] << " ";
	else if(toks[i].type == BLOCKOPEN)
	    o << "{";
	else if(toks[i].type == BLOCKCLOSE)
	    o << "}";
	else if(toks[i].type == VALUE)
	    o << toks[i].var << " ";
	else if(toks[i].type == VARNAME)
	    o << ">" << toks[i].name << "< ";
	else if(toks[i].type == STRING)
	    o << toks[i].var << " ";
	else if(toks[i].type == FUNCCALL ||
		toks[i].type == IF ||
		toks[i].type == FOR ||
		toks[i].type == WHILE)
	{
	    if(toks[i].type == FUNCCALL)
		o << toks[i].name << "( ";
	    else if(toks[i].type == IF)
		o << "if( ";
	    else if(toks[i].type == FOR)
		o << "for( ";
	    else if(toks[i].type == WHILE)
		o << "while( ";
	    for(size_t j=0; j < toks[i].parameters.size(); j++)
	    {
		if(toks[j].type == NOP)
		{
		    if(j+1 < toks[i].parameters.size() &&
		       toks[i].type == FUNCCALL)
			o << " , ";
		    else if(j+1 < toks[i].parameters.size())
			o << " ; ";
		}
		else
		    o << toks[i].parameters;
	    }
	    o << ") ";
	}
	else if(toks[i].type == FUNCDEF)
	{
	    o << toks[i].name << "( ";
	    for(size_t j=0; j < toks[i].parameters.size(); j++)
	    {
		if(toks[i].type == NOP)
		{
		    if(j+1 < toks[i].parameters.size())
			o << " , ";
		}
		else
		    o << toks[i].parameters;
	    }
	    o << ")";
	}
	else if(toks[i].type == NOP)
	    o << "#";
	else if(toks[i].type == EOL)
	    o << ";" << endl;
    }
    ENDRETURN(o);
}
