#include "debug.h"

/*!
 * \file   debug.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:47:41 2006
 * 
 * \brief  Functions for debugging and profiling code
 */

list <FuncCall> FunctionCalls;
bool debugging = false;

void ProfileBegin(const string funcname)
{
    //! \todo begin profiling
    return;
}

void ProfileEnd()
{
    //! \todo end profiling
    return;
}

void BeginDebug()
{
    debugging = true;
}

void FuncEntry(const string funcname)
{
    // this is to stop statically initialised functions from screwing things up
    if(debugging)
    {
	FuncCall call;
	call.name = funcname;
	FunctionCalls.push_back(call);
    }
}

void FuncLeave(const string funcname)
{
    if(debugging)
    {
	FuncCall call = FunctionCalls.back();
	assert(call.name == funcname);
	FunctionCalls.pop_back();
    }
}

bool StackTrace()
{
    int i = FunctionCalls.size();
    for(list<FuncCall>::iterator it = FunctionCalls.end();
	it != FunctionCalls.begin();)
    {
	--it;
	Log << i << " -> " << (*it).name << endl;
	--i;
    }
    return true;
}

void handler(int s)
{
    struct sig_info
    {
	sig_info(int i, char *s, char *l)
	{
	    signumber = i;
	    memset(shortname, 0, sizeof(shortname));
	    strncpy(shortname, s, sizeof(shortname)-1);
	    memset(longname, 0, sizeof(longname));
	    strncpy(longname, l, sizeof(longname)-1);
	}
	int signumber;
	char shortname[10];
	char longname[64];
    };
    const sig_info info[] = {
	sig_info(1, "SIGHUP",	"Hangup (POSIX)"),
	sig_info(2, "SIGINT",	"Terminal interrupt (ANSI)"),
	sig_info(3, "SIGQUIT",	"Terminal quit (POSIX)"),
	sig_info(4, "SIGILL",	"Illegal instruction (ANSI)"),
	sig_info(5, "SIGTRAP",	"Trace trap (POSIX)"),
	sig_info(6, "SIGIOT",	"IOT Trap (4.2 BSD)"),
	sig_info(7, "SIGBUS",	"BUS error (4.2 BSD)"),
	sig_info(8, "SIGFPE",	"Floating point exception (ANSI)"),
	sig_info(9, "SIGKILL",	"Kill(can't be caught or ignored) (POSIX)"),
	sig_info(10, "SIGUSR1",	"User defined signal 1 (POSIX)"),
	sig_info(11, "SIGSEGV",	"Invalid memory segment access (ANSI)"),
	sig_info(12, "SIGUSR2",	"User defined signal 2 (POSIX)"),
	sig_info(13, "SIGPIPE",	"Write on a pipe with no reader, Broken pipe (POSIX)"),
	sig_info(14, "SIGALRM",	"Alarm clock (POSIX)"),
	sig_info(15, "SIGTERM",	"Termination (ANSI)"),
	sig_info(16, "SIGSTKFLT","Stack fault"),
	sig_info(17, "SIGCHLD",	"Child process has stopped or exited, changed (POSIX)"),
	sig_info(18, "SIGCONT",	"Continue executing, if stopped (POSIX)"),
	sig_info(19, "SIGSTOP",	"Stop executing(can't be caught or ignored) (POSIX)"),
	sig_info(20, "SIGTSTP",	"Terminal stop signal (POSIX)"),
	sig_info(21, "SIGTTIN",	"Background process trying to read, from TTY (POSIX)"),
	sig_info(22, "SIGTTOU",	"Background process trying to write, to TTY (POSIX)"),
	sig_info(23, "SIGURG",	"Urgent condition on socket (4.2 BSD)"),
	sig_info(24, "SIGXCPU",	"CPU limit exceeded (4.2 BSD)"),
	sig_info(25, "SIGXFSZ",	"File size limit exceeded (4.2 BSD)"),
	sig_info(26, "SIGVTALRM","Virtual alarm clock (4.2 BSD)"),
	sig_info(27, "SIGPROF",	"Profiling alarm clock (4.2 BSD)"),
	sig_info(28, "SIGWINCH","Window size change (4.3 BSD, Sun)"),
	sig_info(29, "SIGIO",	"I/O now possible (4.2 BSD)"),
	sig_info(30, "SIGPWR",	"Power failure restart (System V)")
    };

    for(size_t i=0; i < sizeof(info)/sizeof(sig_info); i++)
    {
	if(s == info[i].signumber)
	    Log << "Caught signal " << info[i].shortname << " - " << info[i].longname << endl;
    }
    Log << "Stack Trace:" << endl;
    STACK();
    exit(1);
}

bool AddSignalHandlers()
{
#ifndef _WIN32
    signal(SIGQUIT, handler);
    signal(SIGHUP, handler);
#endif
    signal(SIGINT, handler);
    signal(SIGILL, handler);
    signal(SIGABRT, handler);
    signal(SIGSEGV, handler);
    return true;
}

#include "memory.h"

void Assert(string expr, bool value)
{
    if(value)
	return;

    Log << "Assert failed: " << expr << endl;
    STACK();
    exit(0);
}
