#ifndef FONT_H
#define FONT_H

#include "general.h"
#include "matrix.h"
#include "texture.h"
#include "log.h"

#include <ft2build.h>
#include <freetype/ftglyph.h>
#include FT_FREETYPE_H

/*!
 * \file   font.h
 * \author  <baldurk@baldurk.org>
 * \date   Fri Jan  6 16:24:12 2006
 *
 * \brief  A class for rendering fonts via freetype  
 */

/*!
 * \brief A class that renders fonts
 * 
 */
class FontRender
{
  public:
    //! A default constructor, setting the initial scale to 1:1
    FontRender() : xsize(1.0f), ysize(1.0f) {FUNC("FontRender::FontRender");END();}
    /*!
     * \brief This initialises the class to use a specified ttf file
     * 
     * \param fontname the path to the ttf file
     */

    bool Init(string fontname);
    /*!
     * \brief This prints a formatted string in printf style
     * 
     * \param str the formatting string
     * \param ... any other variables
     */

    bool Print(const char *str, ...);

    /*!
     * \brief Sets the size of the printed string
     * 
     * Values above 1.0 will possibly be visibly low detail.
     *
     * \param x the scale in the x direction
     * \param y the scale in the y direction
     */
    inline bool SetSize(float x, float y)
    { FUNC("FontRender::SetSize");xsize = x; ysize = y; ENDRETURN(true); }

    /*!
     * \brief This gets the length of a formatted string in printf style
     * 
     * \param str the formatting string
     * \param ... any other variables
     */

    float Length(const char *str, ...);
  private:
    //! The current x size
    float xsize,
	//! the current y size
	ysize;
    //! the handle on the display lists
    GLuint chars;
    float lengths[128];
};

#endif

