#include "memory.h"

/*!
 * \file   general.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:44:58 2006
 * 
 * \brief  General functions of use
 */

list<MMObject *> MMObject::objects;
bool MMObject::shutdown = false;

bool fleq(const float a, const float b)
{
    const float epsilon = 0.008f;

    if(fabs(a-b) < epsilon)
	return true;

    return false;
}

template <class T> Vector<T> RayPlaneIntersect(Vector <T> planenormal,
					       Vector <T> vecbegin, Vector <T> vecdir,
					       Vector <T> planepos)
{
    FUNC("RayPlaneIntersect");
    T t = ( (planenormal ^ planepos) - (planenormal ^ vecbegin) )
	/
	(planenormal ^ vecdir);

    ENDRETURN(vecdir * t + vecbegin);
}

template <class T> bool PointOnTriangle(Vector <T> pnt, const Vector <T> *tri)
{
    FUNC("PointOnTriangle, tri*");
    Vector <T> one = -pnt + tri[0];
    Vector <T> two = -pnt + tri[1];
    Vector <T> three = -pnt + tri[2];
    one.normalise();two.normalise();three.normalise();

    float ang = acosf(one ^ two) + acosf(two ^ three) + acosf(three ^ one);
    if(fabs(ang - (2 * PI)) < 0.1f)
	ENDRETURN(true);

    ENDRETURN(false);
}

template <class T> bool PointOnTriangle(Vector <T> pnt,
					const Vector <T> tri1,
					const Vector <T> tri2,
					const Vector <T> tri3)
{
    FUNC("PointOnTriangle, tri123");
    Vector <T> one = -pnt + tri1;
    Vector <T> two = -pnt + tri2;
    Vector <T> three = -pnt + tri3;
    one.normalise();
    two.normalise();
    three.normalise();

    float ang = acosf(one ^ two) + acosf(two ^ three) + acosf(three ^ one);
    if(fabs(ang - (2 * PI)) < 0.1f)
	ENDRETURN(true);

    ENDRETURN(false);
}

string ReadFile(string filename)
{
    FUNC("ReadFile");
    string ret;
    string line;
    ifstream myfile(filename.c_str());
    if(myfile.is_open())
    {
	while(!myfile.eof())
	{
	    getline(myfile, line);
	    ret += line + "\n";
	}
	myfile.close();
    }
    ENDRETURN(ret);
}

int substr_count(string haystack, string needle)
{
    int count = 0;
    int last_pos = 0;
    while(haystack.find(needle, last_pos) != string::npos)
    {
	count++;
	last_pos = haystack.find(needle, last_pos)+needle.length();
    }
    return count;
}

string str_replace(string from, string to, string conv)
{
    string ret;
    while(conv.find(from) != string::npos)
    {
	ret += conv.substr(0, conv.find(from));
	ret += to;
	conv = conv.substr(conv.find(from)+from.length());
    }
    ret += conv;
    return ret;
}
