#ifndef GENERAL_H
#define GENERAL_H

#include <cstring>

#include <iostream>
#include <sstream>
#include <fstream>
#include <algorithm>
#include <string>

#ifndef NO_LOG
#include "log.h"
#endif

#ifndef NO_STL
#include <set>
#include <map>
#include <list>
#include <vector>
#endif

using namespace std;

/*!
 * \file   general.h
 * \author  <baldurk@baldurk.org>
 * \date   Fri Dec 30 21:46:59 2005
 * 
 * \brief  Generals definitions and functions
 */

/*! \mainpage notitle
 *
 * \section introsec Introduction
 * This is the doxygen generated documentation for the BK 'Engine' mk.II.
 * 
 * This is my first time using doxygen on a real proper application
 */

#ifndef NO_DEBUG
#include "debug.h"
#endif

#ifdef _WIN32
#define vsnprintf _vsnprintf
#endif

#ifdef _WIN32
#define round ceil
#endif

#ifndef NO_MATH
#include <cmath>

/*! 
 * returns a random number between min and max
 *
 * \param min the lowest number in the bound
 * \param max the highest number in the bound
 */
#define RAND(min, max) ((float(rand())/float(RAND_MAX)*(float(max)-float(min)))+float(min))

//! floating point comparison
bool fleq(const float a, const float b);

//! the famous constant
#define PI 3.1415926535f

#ifndef M_PI
#define M_PI PI
#endif

//! convert degrees to radians
#define DEGTORAD(a) (a*(PI/180.0f))
//! convert radians to degrees
#define RADTODEG(a) (a*(180.0f/PI))

#include "vec.h"
#include "matrix.h"

/*! 
 * \brief Finds the intersection between a ray and a plane, given the plane normal,
 * a point on the plane and the start and direction of the ray.
 *
 * \param planenormal the normal of the plane in question
 * \param vecbegin    the point the ray starts at
 * \param vecdir      a normal pointing in the direction of the ray
 * \param planepos    a point on the plane
 */
template <class T> Vector<T> RayPlaneIntersect(Vector <T> planenormal,
					       Vector <T> vecbegin, Vector <T> vecdir,
					       Vector <T> planepos);

/*! 
 * \brief Finds out if a point lies upon a triangle
 *
 * \param pnt the point
 * \param tri an array of points that specifies the vertices of the triangle
 */
template <class T> bool PointOnTriangle(Vector <T> pnt, const Vector <T> *tri);

/*! 
 * \brief Finds out if a point lies upon a triangle
 *
 * \param pnt  the point
 * \param tri1 the first vertex on triangle
 * \param tri2 the second vertex
 * \param tri3 the third vertex
 */
template <class T> bool PointOnTriangle(Vector <T> pnt,
					const Vector <T> tri1,
					const Vector <T> tri2,
					const Vector <T> tri3);

/*!
 * \brief Reads the specified file into a string
 * 
 * \param filename the filename of the file to load
 */
string ReadFile(string filename);

/*!
 * \brief Returns the number of times the needle appears in the haystack
 * 
 * \param haystack the string to search
 * \param needle   the string to search for
 */
int substr_count(string haystack, string needle);

/*!
 * \brief Replaces one arbitrary string with another
 * 
 * \param from the string to go from
 * \param to   the string to go to
 * \param conv the string to convert
 */

string str_replace(string from, string to, string conv);

#endif

#endif
