#include "glsl.h"

/*!
 * \file   glsl.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:50:56 2006
 * 
 * \brief  GLSL helper classes
 */

map<string, GLhandleARB> GLSL::frags, GLSL::verts;

bool GLSL::AddFragShader(string filename)
{
    FUNC("GLSL::AddFragShader");

    frag = filename;
    if(frags[frag] == 0)
    {
	frags[frag] = glCreateShader(GL_FRAGMENT_SHADER);

	string fs = ReadFile(filename);
	char *ff = new char[fs.length()+1];

	strncpy(ff, fs.c_str(), fs.length());
	ff[fs.length()] = 0;

	glShaderSource(frags[frag], 1, (const char **)&ff, NULL);

	delete[] ff;

	glCompileShader(frags[frag]);

	int vertCompiled = 9;
	glGetShaderiv(verts[vert], GL_COMPILE_STATUS, &vertCompiled);
    }

    ENDRETURN(true);
}

bool GLSL::AddVertShader(string filename)
{
    FUNC("GLSL::AddVertShader");

    vert = filename;
    if(verts[vert] == 0)
    {
	verts[vert] = glCreateShader(GL_VERTEX_SHADER);

	string vs = ReadFile(filename);
	char *vv = new char[vs.length()+1];

	strncpy(vv, vs.c_str(), vs.length());
	vv[vs.length()] = 0;
	glShaderSource(verts[vert], 1, (const char **)&vv, NULL);

	delete[] vv;

	glCompileShader(verts[vert]);

	int vertCompiled = 9;
	glGetShaderiv(verts[vert], GL_COMPILE_STATUS, &vertCompiled);
    }

    ENDRETURN(true);
}

bool GLSL::Compile()
{
    FUNC("GLSL::Compile");

    if(vert != "" && frag != "")
    {
	prog = glCreateProgram();

	glAttachShader(prog, verts[vert]);
	glAttachShader(prog, frags[frag]);

	glLinkProgram(prog);

	int linked = 9;
	glGetProgramiv(prog, GL_LINK_STATUS, &linked);
    }

    ENDRETURN(true);
}

bool GLSL::Bind()
{
    FUNC("GLSL::Bind");

    if(prog != 0)
	glUseProgram(prog);

    ENDRETURN(true);
}

void GLSL::Uniform1f(string name, float a)
{
    int loc = glGetUniformLocation(prog, name.c_str());
    if(loc == -1)
	Log << "no such uniform: " << name << endl;
    else
	glUniform1f(loc, a);
}

void GLSL::Uniform2f(string name, float a, float b)
{
    int loc = glGetUniformLocation(prog, name.c_str());
    if(loc == -1)
	Log << "no such uniform: " << name << endl;
    else
	glUniform2f(loc, a, b);
}

void GLSL::Uniform3f(string name, float a, float b, float c)
{
    int loc = glGetUniformLocation(prog, name.c_str());
    if(loc == -1)
	Log << "no such uniform: " << name << endl;
    else
	glUniform3f(loc, a, b, c);
}


