#ifndef GLSL_H
#define GLSL_H

/*!
 * \file   glsl.h
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:49:09 2006
 * 
 * \brief  GLSL helper classes
 */

#include "general.h"
#include "GLee.h"

//! a class to load up GLSL programs and bind them
class GLSL
{
  public:
    //! a default constructor
    GLSL() : vert(""), frag(""), prog(0) {}

    /*!
     * load a fragment shader into this program
     * 
     * \param filename the filename of the shader to load
     */
    bool AddFragShader(string filename);

    /*!
     * load a vertex shader into this program
     * 
     * \param filename the filename of the shader to load
     */
    bool AddVertShader(string filename);

    /*!
     * Set a uniform variable
     * 
     * \param name the name of the uniform
     * \param a    the first parameter
     * \param b    the second parameter
     * \param c    the third parameter
     */
    void Uniform3f(string name, float a, float b, float c);

    /*!
     * Set a uniform variable
     * 
     * \param name the name of the uniform
     * \param a    the first parameter
     * \param b    the second parameter
     */
    void Uniform2f(string name, float a, float b);

    /*!
     * Set a uniform variable
     * 
     * \param name the name of the uniform
     * \param a    the first parameter
     */
    void Uniform1f(string name, float a);

    //! Compile the shaders into a program
    bool Compile();

    //! Bind the program to the current pipeline
    bool Bind();
  private:
    //! a static map across all GLSL programs to save the handles
    //! of loaded fragment shaders
    static map<string, GLhandleARB> frags,
	//! a static map across all GLSL programs to save the handles
	//! of loaded vertex shaders
	verts;
    //! the current filename of the vertex shader
    string vert,
	//! the current filename of the vertex shader
	frag;
    //! the handle of the program
    GLhandleARB prog;
};

#endif
