#include "glwidget.h"

bool Combobox::MouseClick(bool left, bool middle, bool right)
{
    FUNC("Combobox::MouseClick");
    int f = -1;
    for(size_t i=0; i < children.size(); i++)
	if(children[i]->MouseClick(left, middle, right) && children[i]->Unique())
	    f = i;
    if(f > -1)
    {
	for(size_t i=0; i < children.size(); i++)
	    if((signed)i != f)
		children[i]->SetFocussed(children[f]);
    }

    if(!prevclick && (children[0]->Clicked() || 
		      (left && mpos.x > pos.x &&
		       mpos.x < pos.x+dim.x &&
		       mpos.y > pos.y && mpos.y < pos.y+dim.y)) && !lastleft)
    {
	prevclick = true;
	menu = !menu;
	left = false;
    }
    else if(!children[0]->Clicked() &&
	    !(left && mpos.x > pos.x &&
	      mpos.x < pos.x+dim.x &&
	      mpos.y > pos.y && mpos.y < pos.y+dim.y))
	prevclick = false;

    if(menu &&
       mpos.x > pos.x && mpos.x < pos.x+dim.x &&
       mpos.y < pos.y && mpos.y > pos.y - Texts.size() * theme->textsize.y*2.0f)
    {
	float dy = pos.y - mpos.y;
	int idx = int(dy/(Texts.size() * theme->textsize.y*2.0f)*Texts.size());
	if((left && !lastleft) || (lastleft && !left))
	{
	    SetCurrentText(idx);
	    menu = false;
	}
	else
	    highlightidx = idx;
    }
    else
	highlightidx = -1;

    /*   
    if(left && !(mpos.x > pos.x &&
		 mpos.x < pos.x+dim.x &&
		 mpos.y > pos.y && mpos.y < pos.y+dim.y))
    {
	menu = false;
    }
    */

    if(menu)
	z = -1.0f;
    else
	z = 0.0f;

    lastleft = left;
    lastmiddle = middle;
    lastright = right;
    ENDRETURN(true);
}

bool Combobox::MouseMove(float x, float y)
{
    FUNC("Combobox::MouseMove");
    mpos = Vectorf(x, y);

    for(size_t i=0; i < children.size(); i++)
	children[i]->MouseMove(x-pos.x, y-pos.y);

    ENDRETURN(true);
}

bool Combobox::Render()
{
    FUNC("GLWindow::Render");
    glPushMatrix();

    glTranslatef(pos.x+dim.x/2, pos.y+dim.y/2, 0.0f);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_TEXTURE_2D);

    glLineWidth(theme->border);
    float dimx = dim.x, dimy = dim.y;
    glEnable(GL_BLEND);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.x,
	      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.y,
	      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.z,
	      theme->mainalpha);

    for(int i=0; i < 2; i++)
    {
	if(i == 1)
	{
	    if(focus && theme->highlight)
	    {
		glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.x,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.y,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.z,
			  theme->mainalpha);
	    }
	    else
	    {
		glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.x,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.y,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.z,
			  theme->mainalpha);
	    }
	    glDisable(GL_TEXTURE_2D);
	    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
	else
	{
	    if(focus && theme->highlight)
	    {
		glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.x,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.y,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.z,
			  theme->mainalpha);
	    }
	    else
	    {
		glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.x,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.y,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.z,
			  theme->mainalpha);
	    }
	    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
	glBegin(GL_POLYGON);
	glVertex2f(-dimx/2, -dimy/2);
	glVertex2f(-dimx/2, -dimy/2+dimy*(1.0f-theme->corner));
	glVertex2f(-dimx/2+dimy*theme->corner, -dimy/2+dimy);
	glVertex2f(-dimx/2+dimx, -dimy/2+dimy);
	glVertex2f(-dimx/2+dimx, -dimy/2+dimy*theme->corner);
	glVertex2f(-dimx/2+dimx-dimy*theme->corner, -dimy/2);
	glVertex2f(-dimx/2, -dimy/2);
	glEnd();
    }
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    if(menu)
    {
	float menuheight = Texts.size() * theme->textsize.y*2.0f;

	for(int i=0; i < 2; i++)
	{
	    if(i == 1)
	    {
		if(focus && theme->highlight)
		{
		    glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.x,
			      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.y,
			      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.z,
			      theme->secondalpha);
		}
		else
		{
		    glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.x,
			      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.y,
			      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.z,
			      theme->secondalpha);
		}
		glDisable(GL_TEXTURE_2D);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	    }
	    else
	    {
		if(focus && theme->highlight)
		{
		    glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.x,
			      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.y,
			      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.z,
			      theme->secondalpha);
		}
		else
		{
		    glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.x,
			      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.y,
			      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.z,
			      theme->secondalpha);
		}
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	    }
	    
	    glPointSize(5.0f);
	    glBegin(GL_POLYGON);
	    glVertex2f(-dimx/2+dimx, -dimy/2);
	    glVertex2f(-dimx/2+dimx, -dimy/2-menuheight);
	    glVertex2f(-dimx/2, -dimy/2-menuheight);
	    glVertex2f(-dimx/2, -dimy/2);
	    glVertex2f(-dimx/2+dimx-dimy*theme->corner, -dimy/2);
	    glVertex2f(-dimx/2+dimx, -dimy/2+dimy*theme->corner);
	    glEnd();
	}
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	for(size_t i=0; i < Texts.size(); i++)
	{
	    glPushMatrix();
	    glTranslatef(children[0]->Getbbox().z+font.Length(" ")-dimx/2+theme->insideborder*3.0f,
			 -dimy/2+theme->textsize.y/2.0f-(i+1)*theme->textsize.y*2.0f, 0.0f);
	    if((int)i == highlightidx)
	    {
		glColor4f(theme->bordercol.x*theme->dimfact,
			  theme->bordercol.y*theme->dimfact,
			  theme->bordercol.z*theme->dimfact,
			  theme->secondalpha);
		glBegin(GL_QUADS);
		glVertex2f(0.0f, -theme->textsize.y/3.0f);
		glVertex2f(0.0f, theme->textsize.y);
		glVertex2f(font.Length(Texts[i].c_str()), theme->textsize.y);
		glVertex2f(font.Length(Texts[i].c_str()), -theme->textsize.y/3.0f);
		glEnd();
	    }
	    glColor4f(theme->textcol.x, theme->textcol.y, theme->textcol.z, 1.0f);
	    font.Print(Texts[i].c_str());
	    glPopMatrix();
	}
    }

    glPopMatrix();

    glPushMatrix();
    glTranslatef(pos.x+theme->insideborder*3.0f, pos.y+theme->insideborder*3.0f, 0.0f);
    for(size_t i=0; i < children.size(); i++)
    {
	glPushMatrix();
	children[i]->Render();
	glPopMatrix();
    }
    glTranslatef(children[0]->Getbbox().z+font.Length(" "),
		 theme->textsize.y/3.0f, 0.0f);
    font.Print(GetCurrentText().c_str());

    glPopMatrix();
    ENDRETURN(true);
}

bool Combobox::AddText(string t)
{
    FUNC("Combobox::AddText");
    GLWidget::AddText(t);
    for(size_t i=0; i < Texts.size(); i++)
	if(dim.x < font.Length(Texts[i].c_str())+theme->textsize.x*2.0f+children[0]->Getbbox().z+theme->insideborder*6.0f)
	    dim.x = font.Length(Texts[i].c_str())+theme->textsize.x*2.0f+children[0]->Getbbox().z+theme->insideborder*6.0f;
    ENDRETURN(true);
}
