#include "glwidget.h"

bool Frame::Render()
{
    FUNC("Frame::Render");
    glPushMatrix();

    glTranslatef(pos.x+dim.x/2, pos.y+dim.y/2, 0.0f);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_TEXTURE_2D);

    glLineWidth(theme->border);
    float dimx = dim.x, dimy = dim.y;

    if(bordered)
    {
	glColor4f(theme->dimfact*theme->bordercol.x,
		  theme->dimfact*theme->bordercol.y,
		  theme->dimfact*theme->bordercol.z,
		  theme->mainalpha);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    
	glBegin(GL_POLYGON);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(-dimx/2, -dimy/2);
	glTexCoord2f(0.0f, (1.0f-theme->corner));
	glVertex2f(-dimx/2, -dimy/2+dimy*(1.0f-theme->corner));
	glTexCoord2f(theme->corner, 1.0f);
	glVertex2f(-dimx/2+dimy*theme->corner, -dimy/2+dimy);
	glTexCoord2f(1.0f, 1.0f);
	glVertex2f(-dimx/2+dimx, -dimy/2+dimy);
	glTexCoord2f(1.0f, theme->corner);
	glVertex2f(-dimx/2+dimx, -dimy/2+dimy*theme->corner);
	glTexCoord2f(1.0f, 0.0f);
	glVertex2f(-dimx/2+dimx-dimy*theme->corner, -dimy/2);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(-dimx/2, -dimy/2);
	glEnd();

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    glDisable(GL_TEXTURE_2D);

    glPopMatrix();

    glPushMatrix();
    if(bordered)
	glTranslatef(pos.x+theme->insideborder*3.0f, pos.y+theme->insideborder*3.0f, 0.0f);
    else
	glTranslatef(pos.x, pos.y, 0.0f);
    for(size_t i=0; i < children.size(); i++)
    {
	glPushMatrix();
	if(children[i]->GetDepth() > -0.5f)
	    children[i]->Render();
	glPopMatrix();
    }
    for(size_t i=0; i < children.size(); i++)
    {
	glPushMatrix();
	if(children[i]->GetDepth() < -0.5f)
	    children[i]->Render();
	glPopMatrix();
    }
    glPopMatrix();
    ENDRETURN(true);
}

bool Frame::MouseClick(bool left, bool middle, bool right)
{
    FUNC("Frame::MouseClick");
    int f = -1;
    for(size_t i=0; i < children.size(); i++)
	if(children[i]->MouseClick(left, middle, right) && children[i]->Unique())
	    f = i;

    if(f > -1)
	for(size_t i=0; i < children.size(); i++)
	    if((signed)i != f)
		children[i]->SetFocussed(children[f]);

    ENDRETURN(true);
}

bool Frame::MouseMove(float x, float y)
{
    FUNC("Frame::MouseMove");
    mpos = Vectorf(x, y);

    if(bordered)
	for(size_t i=0; i < children.size(); i++)
	    children[i]->MouseMove(x-pos.x-theme->insideborder*3.0f, y-pos.y-theme->insideborder*3.0f);
    else
	for(size_t i=0; i < children.size(); i++)
	    children[i]->MouseMove(x-pos.x, y-pos.y);

    ENDRETURN(true);
}

bool Frame::SetFocussed(GLWidget *widget)
{
    FUNC("Frame::SetFocussed");
    for(size_t i=0; i < children.size(); i++)
	children[i]->SetFocussed(widget);
    ENDRETURN(true);
}

