#include "glwidget.h"

int GLWidget::viewport[4];
int GLWidget::msgreturn;
bool GLWidget::rendering;
float GLWidget::timescale;
Vectorf GLWidget::optionbox,
    GLWidget::optiondot,
    GLWidget::checkbox,
    GLWidget::checkdot,
    GLWidget::slider;
map<string, GLWidget*> GLWidget::tags;
map<string, Theme*> GLWidget::themes;
FontRender GLWidget::font;
vector<GLWidget *> GLWidget::roots;

void Theme::Default()
{
    textcol = Vectorf(1.0f, 1.0f, 1.0f);
    bordercol = Vectorf(1.0f, 0.71f, 0.16f);
    maincol = Vectorf(0.8f, 0.8f, 0.8f);
    secondcol = Vectorf(0.3f, 0.3f, 0.3f);

    mainalpha = 1.0f; secondalpha = 0.8f;

    border = 2.0f; insideborder = 0.005f;

    corner = 0.1f; dimspeed = 0.003f;

    textsize = Vectorf(0.04f, 0.05f);

    dimfact = 0.7f; brightfact = 1.0f;

    clickoffset = Vectorf(0.005f, -0.005f);

    itemsize = 0.1f;

    tex.Load("gui.tga");
    highlight = true;
}

void Theme::Read(string file)
{
    FUNC("Theme::Read");
    ifstream ifs(file.c_str());
    Log << "reading theme " << file << endl;
    ifs >> textcol.x >> textcol.y >> textcol.z;ifs.ignore(10000,'\n');
    Log << textcol << endl;
    ifs >> bordercol.x >> bordercol.y >> bordercol.z;ifs.ignore(10000,'\n');
    Log << bordercol << endl;
    ifs >> maincol.x >> maincol.y >> maincol.z;ifs.ignore(10000,'\n');
    Log << maincol << endl;
    ifs >> secondcol.x >> secondcol.y >> secondcol.z;ifs.ignore(10000,'\n');
    Log << secondcol << endl;

    ifs >> mainalpha;ifs.ignore(10000,'\n');
    Log << mainalpha << endl;
    ifs >> secondalpha;ifs.ignore(10000,'\n');
    Log << secondalpha << endl;

    ifs >> border;ifs.ignore(10000,'\n');
    Log << border << endl;
    ifs >> insideborder;ifs.ignore(10000,'\n');
    Log << insideborder << endl;

    ifs >> corner;ifs.ignore(10000,'\n');
    Log << corner << endl;
    ifs >> dimspeed;ifs.ignore(10000,'\n');
    Log << dimspeed << endl;

    ifs >> textsize.x >> textsize.y;ifs.ignore(10000,'\n');
    Log << textsize << endl;
    ifs >> dimfact >> brightfact;ifs.ignore(10000,'\n');
    Log << dimfact << " " << brightfact << endl;

    ifs >> clickoffset.x >> clickoffset.y >> clickoffset.z;ifs.ignore(10000,'\n');
    Log << clickoffset << endl;

    ifs >> itemsize;ifs.ignore(10000,'\n');
    Log << itemsize << endl;

    ifs >> texture;ifs.ignore(10000,'\n');
    Log << texture << endl;

    ifs >> highlight;ifs.ignore(10000,'\n');
    Log << highlight << endl;

    if(string(texture) != "NULL")
    {
	Log << "loading " << string(texture) << endl;
	tex.Load(texture);
    }

    Log << "read theme " << file << endl;

    ifs.close();
    END();
}

bool GLWidget::StartGUI(string fname)
{
    FUNC("GLWidget::StartGUI");
    ifstream ifs(fname.c_str());
    vector<GLWidget *> parents, windows;
    vector<int> tabbed;
    Theme *theme;
    FILE *fp = fopen("default.theme", "r");
    if(!fp)
    {
	themes["default"] = new Theme;
	themes["default"]->Default();
    }
    else
    {
	themes["default"] = new Theme;
	themes["default"]->Read("default.theme");
	fclose(fp);
    }

    bool solid = false;
    while(!ifs.eof())
    {
	string cmd;
	ifs >> cmd;
	if(cmd.substr(0, 2) == "//");
	else if(cmd == "CONF")
	{
	    string file;
	    ifs >> file;
	    Init(file);
	}
	else if(cmd == "CADD" || cmd == "TADD" || cmd == "CHLD")
	{
	    int num;
	    bool drag, bordered;
	    string type, tag, par, title;
	    char rest[512];
	    Vectorf pos, end, dim;
	    if(cmd == "TADD")
		ifs >> num >> par >> type >> tag >> pos.x >> pos.y;
	    else if(cmd == "CADD")
		ifs >> type >> tag >> par >> pos.x >> pos.y;

	    GLWidget *parent;

	    if(par == "NULL")
		parent = NULL;
	    else
		parent = GLWidget::GetByTag(par);

	    Log << "adding a '" << type << "' - " << tag << endl;

	    if(type == "Win")
	    {
		ifs >> drag;
		ifs.get(rest, 512, '\n');
		title = rest;
		while(title.length() && isspace(title[0]))
		    title = title.substr(1);

		Log << "adding window " << tag << " " << drag << endl;

		if(parent == NULL)
		{
		    roots.push_back(new GLWindow(tag, NULL,
						 pos, theme,
						 title, drag));
		}
		else
		{
		    windows.push_back(new GLWindow(tag, parent,
						   pos, theme,
						   title, drag));
		    parents.push_back(parent);
		    if(cmd == "CADD")
			tabbed.push_back(-1);
		    else if(cmd == "TADD")
			tabbed.push_back(num);
		}
	    }
	    if(type == "Frame")
	    {
		ifs >> bordered;

		if(parent == NULL)
		{
		    roots.push_back(new Frame(tag, NULL,
					     pos, theme,
					     bordered));
		}
		else
		{
		    windows.push_back(new Frame(tag, parent,
						pos, theme,
						bordered));
		    parents.push_back(parent);
		    if(cmd == "CADD")
			tabbed.push_back(-1);
		    else if(cmd == "TADD")
			tabbed.push_back(num);
		}
	    }
	    if(type == "Check")
	    {
		ifs.get(rest, 512, '\n');
		title = rest;
		while((int)title.length() > 0 && isspace(title[0]))
		    title = title.substr(1);

		if(parent == NULL)
		    roots.push_back(new Checkbox(tag, NULL,
						pos, theme,
						title));
		else if(cmd == "CADD")
		    parent->AddChild(new Checkbox(tag, parent,
						  pos, theme,
						  title));
		else if(cmd == "TADD")
		    parent->AddToTab(num, new Checkbox(tag, parent,
						       pos, theme,
						       title));
	    }
	    if(type == "Opt")
	    {
		ifs.get(rest, 512, '\n');
		title = rest;
		while((int)title.length() > 0 && isspace(title[0]))
		    title = title.substr(1);

		if(parent == NULL)
		    roots.push_back(new Optionbox(tag, NULL,
						 pos, theme,
						 title));
		else if(cmd == "CADD")
		    parent->AddChild(new Optionbox(tag, parent,
						   pos, theme,
						   title));
		else if(cmd == "TADD")
		    parent->AddToTab(num, new Optionbox(tag, parent,
							pos, theme,
							title));
	    }
	    if(type == "Combo")
	    {
		if(parent == NULL)
		    roots.push_back(new Combobox(tag, NULL,
						pos, theme));
		else if(cmd == "CADD")
		    parent->AddChild(new Combobox(tag, parent,
						   pos, theme));
		else if(cmd == "TADD")
		    parent->AddToTab(num, new Combobox(tag, parent,
							pos, theme));
	    }
	    if(type == "Butt")
	    {
		string start;
		ifs >> start;
		if(start == "TEX")
		{
		    ifs >> dim.x >> dim.y;

		    if(parent == NULL)
			roots.push_back(new Button(tag, NULL, pos,
						  theme, dim));
		    else if(cmd == "CADD")
			parent->AddChild(new Button(tag, parent, pos,
						    theme, dim));
		    else if(cmd == "TADD")
			parent->AddToTab(num, new Button(tag, parent, pos,
							 theme, dim));
		}
		else
		{
		    title = start;
		    while(title.length() && isspace(title[0]))
			title = title.substr(1);

		    if(parent == NULL)
			roots.push_back(new Button(tag, NULL, pos,
						  theme, title));
		    else if(cmd == "CADD")
			parent->AddChild(new Button(tag, parent, pos,
						    theme, title));
		    else if(cmd == "TADD")
			parent->AddToTab(num, new Button(tag, parent, pos,
							 theme, title));
		}
	    }
	    if(type == "Slide")
	    {
		ifs >> end.x >> end.y;

		if(parent == NULL)
		    roots.push_back(new Slider(tag, NULL, pos, theme,
					      end));
		else if(cmd == "CADD")
		    parent->AddChild(new Slider(tag, parent, pos, theme,
						end));
		else if(cmd == "TADD")
		    parent->AddToTab(num, new Slider(tag, parent, pos, theme,
						     end));
	    }
	    if(type == "Text")
	    {
		int width, height;
		ifs >> width >> height;

		if(parent == NULL)
		    roots.push_back(new Textbox(tag, NULL, pos, theme,
						width, height));
		else if(cmd == "CADD")
		    parent->AddChild(new Textbox(tag, parent, pos, theme,
						 width, height));
		else if(cmd == "TADD")
		    parent->AddToTab(num, new Textbox(tag, parent, pos, theme,
						      width, height));
	    }
	    if(type == "Label")
	    {
		ifs.get(rest, 512, '\n');
		title = rest;
		while(title.length() && isspace(title[0]))
		    title = title.substr(1);

		title = str_replace("\\n", "\n", title);

		if(parent == NULL)
		    roots.push_back(new Label(tag, NULL, pos, theme,
					     title));
		else if(cmd == "CADD")
		    parent->AddChild(new Label(tag, parent, pos, theme,
						title));
		else if(cmd == "TADD")
		    parent->AddToTab(num, new Label(tag, parent, pos, theme,
						     title));
	    }
	    if(type == "TWin")
	    {
		Log << "adding tabwindow " << tag << " " << parent << " " << pos << " " << theme << endl;

		if(parent == NULL)
		{
		    roots.push_back(new TabWindow(tag, NULL, pos,
						 theme));
		}
		else
		{
		    windows.push_back(new TabWindow(tag, parent, pos,
						    theme));
		    parents.push_back(parent);
		    if(cmd == "CADD")
			tabbed.push_back(-1);
		    else if(cmd == "TADD")
			tabbed.push_back(num);
		}

		Log << "added tabwindow" << endl;
	    }
	}
	else if(cmd == "NTAB")
	{
	    string par, title;
	    int num;
	    ifs >> par >> num;
	    GLWidget *parent;
	    if(par == "NULL")
		parent = NULL;
	    else
		parent = GLWidget::GetByTag(par);
	    char rest[512];
	    ifs.get(rest, 512, '\n');
	    title = rest;
	    while(title.length() && isspace(title[0]))
		title = title.substr(1);

	    parent->AddTab(num, title);
	}
	else if(cmd == "TEXT")
	{
	    string toadd, text;
	    ifs >> toadd;
	    if(toadd != "NULL")
	    {
		GLWidget *addto = GLWidget::GetByTag(toadd);
		char rest[512];
		ifs.get(rest, 512, '\n');
		text = rest;
		while(text.length() && isspace(text[0]))
		    text = text.substr(1);

		text = str_replace("\\n", "\n", text);
		
		addto->AddText(text);
	    }
	}
	else if(cmd == "SOLID")
	{
	    for(size_t i=0; i < parents.size(); i++)
	    {
		if(tabbed[i] != -1)
		    parents[i]->AddToTab(tabbed[i], windows[i]);
		else
		    parents[i]->AddChild(windows[i]);
	    }

	    for(size_t i=0; i < roots.size(); i++)
		roots[i]->Solidify();
	    solid = true;
	}
	else if(cmd == "CENTREX")
	{
	    string title;ifs >> title;
	    GLWidget *wid = GLWidget::GetByTag(title);
	    if(!wid);
	    else if(wid->parent)
		wid->parent->Centre(wid, true, false);
	    else
	    {
		Vectorf vec = wid->Getbbox();
		(*wid) += (Vectorf(-(vec.z-vec.x)/2.0f));
	    }
	}
	else if(cmd == "CENTREY")
	{
	    string title;ifs >> title;
	    GLWidget *wid = GLWidget::GetByTag(title);
	    if(!wid);
	    else if(wid->parent)
		wid->parent->Centre(wid, false, true);
	    else
	    {
		Vectorf vec = wid->Getbbox();
		(*wid) += (Vectorf(0.0f,
				   -(vec.w-vec.y)/2.0f));
	    }
	}
	else if(cmd == "CENTRE")
	{
	    string title;ifs >> title;
	    GLWidget *wid = GLWidget::GetByTag(title);
	    if(!wid);
	    else if(wid->parent)
		wid->parent->Centre(wid);
	    else
	    {
		Vectorf vec = wid->Getbbox();
		(*wid) += (Vectorf(-(vec.z-vec.x)/2.0f,
				   -(vec.w-vec.y)/2.0f));
	    }
	}
	else if(cmd == "THEM")
	{
	    string filename, additional;ifs >> filename;
	    ifs >> additional;
	    assert(additional == "NULL");

	    if(themes[filename] == NULL)
	    {
		themes[filename] = new Theme;
		theme = themes[filename];
		theme->Read(filename);
	    }
	    else
		theme = themes[filename];
	}
	ifs.ignore(10000,'\n');
    }

    if(!solid)
    {
	for(size_t i=0; i < parents.size(); i++)
	{
	    if(tabbed[i] != -1)
		parents[i]->AddToTab(tabbed[i], windows[i]);
	    else
		parents[i]->AddChild(windows[i]);
	}

	for(size_t i=0; i < roots.size(); i++)
	    roots[i]->Solidify();
    }
    ifs.close();
    ENDRETURN(true);
}

bool GLWidget::Quit()
{
    FUNC("GLWidget::Quit");
    ENDRETURN(true);
}

bool GLWidget::Init(string conffile)
{
    FUNC("GLWidget::Init");
    ifstream ifs(conffile.c_str());
    string fname;
    ifs >> fname;ifs.ignore(10000,'\n');
    Log << "Loading " << fname << endl;
    font.Init(fname.c_str());
    Log << "Loaded." << endl;
    ifs >> fname;ifs.ignore(10000,'\n');
    ifs >> optionbox.x >> optionbox.y >> optionbox.z >> optionbox.w;ifs.ignore(10000,'\n');
    ifs >> optiondot.x >> optiondot.y >> optiondot.z >> optiondot.w;ifs.ignore(10000,'\n');
    ifs >> checkbox.x >> checkbox.y >> checkbox.z >> checkbox.w;ifs.ignore(10000,'\n');
    ifs >> checkdot.x >> checkdot.y >> checkdot.z >> checkdot.w;ifs.ignore(10000,'\n');
    ifs >> slider.x >> slider.y >> slider.z >> slider.w;ifs.ignore(10000,'\n');
    ifs >> timescale;ifs.ignore(10000,'\n');
    ifs.close();
    ENDRETURN(true);
}

bool GLWidget::MessageBox(string message, int buttons)
{
    FUNC("GLWidget::MessageBox");
    themes["messagebox"] = new Theme;
    themes["messagebox"]->Read("messagebox.theme");

    if(!GetByTag("__msgbox"))
    {
	roots.push_back(new GLWindow("__msgbox", NULL,
				     Vectorf(0.0f, 0.0f), themes["messagebox"],
				     "Message box", true));
	GETWIDGET(__msgbox)->AddChild(new Label("__msgbox_message", GetByTag("__msgbox"),
						 Vectorf(0.0f, 0.2f), themes["messagebox"],
						 message));
	if(buttons == OK_BUTTON)
	    GETWIDGET(__msgbox)->AddChild(new Button("__msgbox_ok", GetByTag("__msgbox"),
						      Vectorf(0.0f, 0.0f), themes["messagebox"],
						      "OK"));
	else
	{
	    GETWIDGET(__msgbox)->AddChild(new Frame("__msgbox_frm", GetByTag("__msgbox"),
						     Vectorf(0.0f, 0.0f), themes["messagebox"]));
	    if(buttons == (OK_BUTTON | CANCEL_BUTTON))
	    {
		GETWIDGET(__msgbox_frm)->AddChild(new Button("__msgbox_ok", GetByTag("__msgbox_frm"),
							      Vectorf(0.0, 0.0f), themes["messagebox"],
							      "OK"));
		GETWIDGET(__msgbox_frm)->AddChild(new Button("__msgbox_cancel", GetByTag("__msgbox_frm"),
							      Vectorf(0.2f, 0.0f), themes["messagebox"],
							      "Cancel"));
	    }
	    else if(buttons == (YES_BUTTON | NO_BUTTON))
	    {
		GETWIDGET(__msgbox_frm)->AddChild(new Button("__msgbox_yes", GetByTag("__msgbox_frm"),
							      Vectorf(0.0f, 0.0f), themes["messagebox"],
							      "Yes"));
		GETWIDGET(__msgbox_frm)->AddChild(new Button("__msgbox_no", GetByTag("__msgbox_frm"),
							      Vectorf(0.2f, 0.0f), themes["messagebox"],
							      "No"));
	    }
	    else if(buttons == (YES_BUTTON | NO_BUTTON | CANCEL_BUTTON))
	    {
		GETWIDGET(__msgbox_frm)->AddChild(new Button("__msgbox_yes", GetByTag("__msgbox_frm"),
							      Vectorf(0.0f, 0.0f), themes["messagebox"],
							      "Yes"));
		GETWIDGET(__msgbox_frm)->AddChild(new Button("__msgbox_no", GetByTag("__msgbox_frm"),
							      Vectorf(0.2f, 0.0f), themes["messagebox"],
							      "No"));
		GETWIDGET(__msgbox_frm)->AddChild(new Button("__msgbox_cancel", GetByTag("__msgbox_frm"),
							      Vectorf(0.4f, 0.0f), themes["messagebox"],
							      "Cancel"));
	    }
	}

	GETWIDGET(__msgbox)->Solidify();

	if(buttons == OK_BUTTON)
	{
	    GETWIDGET(__msgbox)->Centre(GetByTag("__msgbox_ok"), true, false);
	}
	else
	{
	    GETWIDGET(__msgbox)->Centre(GetByTag("__msgbox_frm"), true, false);
	}

	GETWIDGET(__msgbox)->Centre(GetByTag("__msgbox_message"), true, false);
	
	{
	    GLWidget *wid = GetByTag("__msgbox");
	    Vectorf vec = wid->Getbbox();
	    (*wid) += (Vectorf(-(vec.z-vec.x)/2.0f,
			       -(vec.w-vec.y)/2.0f));
	}
    }
    ENDRETURN(true);
}

int GLWidget::GetMessage()
{
    int rc = msgreturn;
    msgreturn = 0;
    return rc;
}

GLWidget::GLWidget(string Tag, GLWidget *Parent, Vectorf Pos,
		   Theme *Theme)
{
    FUNC("GLWidget::GLWidget");
    rendering = false;
    tags[Tag] = this;tag = Tag;
    pos = Pos;mpos = Vectorf();
    theme = Theme;
    if(theme == NULL)
	theme = themes["default"];
    lastleft = lastmiddle = lastright = false;
    parent = Parent;Time = 0;
    focusdt = clickdt = 0.0f;focus = beenclicked = clicked = false;
    font.SetSize(theme->textsize.x, theme->textsize.y);
    currenttext = 0;z = 0.0f;msgreturn = 0;
    glGetIntegerv(GL_VIEWPORT, viewport);
    END();
}

bool GLWidget::Centre(GLWidget *child, bool x, bool y)
{
    FUNC("GLWidget::Centre");

    assert(child != 0);

    float dx = ((Getbbox().z-Getbbox().x-theme->insideborder*6.0f)/2.0f)-((child->Getbbox().z-child->Getbbox().x)/2.0f),
	dy = ((Getbbox().w-Getbbox().y-theme->textsize.y*2.0f-theme->insideborder*6.0f)/2.0f)-((child->Getbbox().w-child->Getbbox().y)/2.0f);

    if(x)
	(*child) += Vectorf(dx, 0.0f);
    if(y)
	(*child) += Vectorf(0.0f, dy);

    ENDRETURN(true);
}

bool GLWidget::MouseMove(Vectorf mp)
{
    FUNC("GLWidget::MouseMove(Vector)");
    mpos = Vectorf((2.0f*mp.x/viewport[2]-1.0f)*float(viewport[2])/viewport[3],
		   -2.0f*mp.y/viewport[3]+1.0f);

    MouseMove(mpos.x, mpos.y);
    ENDRETURN(true);
}

bool GLWidget::TextKeyPress(bool keys[256])
{
    FUNC("GLWidget::TextKeyPress");
    for(size_t i=0; i < children.size(); i++)
	children[i]->TextKeyPress(keys);
    ENDRETURN(true);
}

bool GLWidget::MouseMove(float x, float y)
{
    FUNC("GLWidget::MouseMove(x, y)");

    mpos = Vectorf(x, y);

    if(Within())
	focus = true;
    else
	focus = false;

    for(size_t i=0; i < children.size(); i++)
	children[i]->MouseMove(x, y);
    ENDRETURN(true);
}

bool GLWidget::Within()
{
    FUNC("GLWidget::Within");
    if(mpos.x > Getbbox().x && mpos.x < Getbbox().z &&
       mpos.y > Getbbox().y && mpos.y < Getbbox().w)
	ENDRETURN(true);
    ENDRETURN(false);
}

bool GLWidget::Solidify()
{
    FUNC("GLWidget::Solidify");
    for(size_t i=0; i < children.size(); i++)
	children[i]->Solidify();
    for(size_t i=0; i < children.size(); i++)
    {
	if(children[i]->Getbbox().z+theme->insideborder*6.0f > dim.x)
	    dim.x = children[i]->Getbbox().z+theme->insideborder*6.0f;
	if(children[i]->Getbbox().w+theme->insideborder*6.0f > dim.y)
	    dim.y = children[i]->Getbbox().w+theme->insideborder*6.0f;
    }
    ENDRETURN(true);
}

bool GLWidget::AddChild(GLWidget *a)
{
    FUNC("GLWidget::AddChild");
    assert(a != 0);
    children.push_back(a);
    int uniq = -1;
    for(size_t i=0; i < children.size(); i++)
	if(children[i]->Unique())
	{
	    uniq = (signed)i;
	    break;
	}
    if(uniq > -1)
    {
	children[uniq]->Enable();
	for(size_t i=0; i < children.size(); i++)
	    if(children[i]->Unique() && (signed)i != uniq)
		children[i]->SetFocussed(children[uniq]);
    }

    ENDRETURN(true);
}

bool GLWidget::MouseMoveWidgets(Vectorf v)
{
    FUNC("GLWidget::MouseMoveWidgets");
    for(size_t i=0; i < roots.size(); i++)
	roots[i]->MouseMove(v);
    ENDRETURN(true);
}

bool GLWidget::TextKeyPressWidgets(bool keys[256])
{
    FUNC("GLWidget::TextKeyPressWidgets");
    for(size_t i=0; i < roots.size(); i++)
	roots[i]->TextKeyPress(keys);
    ENDRETURN(true);
}

bool GLWidget::MouseClickWidgets(bool a, bool b, bool c)
{
    FUNC("GLWidget::MouseClickWidgets");
    for(size_t i=0; i < roots.size(); i++)
	roots[i]->MouseClick(a, b, c);
    ENDRETURN(true);
}

bool GLWidget::UpdateWidgets(int time)
{
    FUNC("GLWidget::UpdateWidgets");
    for(size_t i=0; i < roots.size(); i++)
	roots[i]->Update(time);

    bool clicked = false;

    if(GETWIDGET(__msgbox_ok)->BeenClicked())
    {
	clicked = true;
	msgreturn = OK_BUTTON;
    }

    if(GETWIDGET(__msgbox_cancel)->BeenClicked())
    {
	clicked = true;
	msgreturn = CANCEL_BUTTON;
    }

    if(GETWIDGET(__msgbox_yes)->BeenClicked())
    {
	clicked = true;
	msgreturn = YES_BUTTON;
    }

    if(GETWIDGET(__msgbox_no)->BeenClicked())
    {
	clicked = true;
	msgreturn = NO_BUTTON;
    }

    if(clicked)
    {
	vector<vector<GLWidget *>::iterator> toerase;

	for(vector<GLWidget *>::iterator i=roots.begin(); i!=roots.end(); ++i)
	    if(*i == GetByTag("__msgbox"))
		toerase.push_back(i);

	for(vector<vector<GLWidget *>::iterator>::iterator i=toerase.begin(); i!=toerase.end(); ++i)
	{
	    delete **i;
	    roots.erase(*i);
	}

	if(GetByTag("__msgbox_ok"))
	    delete GetByTag("__msgbox_ok");
	if(GetByTag("__msgbox_cancel"))
	    delete GetByTag("__msgbox_cancel");
	if(GetByTag("__msgbox_yes"))
	    delete GetByTag("__msgbox_yes");
	if(GetByTag("__msgbox_no"))
	    delete GetByTag("__msgbox_no");
	delete GetByTag("__msgbox_msg");
	tags["__msgbox_ok"] = NULL;
	tags["__msgbox_cancel"] = NULL;
	tags["__msgbox_yes"] = NULL;
	tags["__msgbox_no"] = NULL;
	tags["__msgbox_msg"] = NULL;
	tags["__msgbox"] = NULL;
    }
    ENDRETURN(true);
}

bool GLWidget::RenderWidgets()
{
    FUNC("GLWidget::BeginRender");
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(-float(viewport[2])/viewport[3], float(viewport[2])/viewport[3],
	    -1, 1, -1.0f, 1.0f);
    glMatrixMode(GL_MODELVIEW);
    glPushAttrib(GL_ENABLE_BIT);
    glDisable(GL_LIGHTING);
    glDisable(GL_COLOR_MATERIAL);
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);
    rendering = true;

    for(size_t i=0; i < roots.size(); i++)
	roots[i]->Render();

    glPopAttrib();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    rendering = false;
    ENDRETURN(true);
}

bool GLWidget::Update(int time)
{
    FUNC("GLWidget::Update");
    Time += time;

    if(!theme->highlight)
	focusdt = 1.0f;
    else if(focus && focusdt < 1.0f)
	focusdt += float(time)*timescale;
    else if(!focus && focusdt > 0.0001f)
	focusdt -= float(time)*timescale;
    else if(focus)
	focusdt = 1.0f;
    else if(!focus)
	focusdt = 0.0f;

    if(clickdt > 0.0001f)
	clickdt -= float(time)*timescale;
    else
	clickdt = 0.0f;

    for(size_t i=0; i < children.size(); i++)
	children[i]->Update(time);

    ENDRETURN(true);
}
