#include "glwidget.h"

bool GLWindow::AddChild(GLWidget *a)
{
    FUNC("GLWindow::AddChild");
    GLWidget::AddChild(a);

    if(dim.x < font.Length(Text.c_str())+theme->textsize.x*2.0f)
	dim.x = font.Length(Text.c_str())+theme->textsize.x*2.0f;

    ENDRETURN(true);
}

bool GLWindow::Solidify()
{
    FUNC("GLWindow::Solidify");
    GLWidget::Solidify();

    if(dim.x < font.Length(Text.c_str())+theme->textsize.x*2.0f)
	dim.x = font.Length(Text.c_str())+theme->textsize.x*2.0f;

    for(size_t i=0; i < children.size(); i++)
    {
	if(children[i]->Getbbox().w+theme->insideborder*6.0f+theme->textsize.y*2.0f > dim.y)
	    dim.y = children[i]->Getbbox().w+theme->insideborder*6.0f+theme->textsize.y*2.0f;
    }

    ENDRETURN(true);
}

bool GLWindow::MouseMove(float x, float y)
{
    FUNC("GLWindow::MouseMove");
    mpos = Vectorf(x, y);

    if(dragging)
    {
	bool drag = false;
	if(!parent)
	    drag = true;

	pos += Vectorf(mpos.x - prevmpos.x,
		       mpos.y - prevmpos.y);
	Vectorf newbbox = Getbbox();
	pos += Vectorf(-(mpos.x - prevmpos.x),
		       -(mpos.y - prevmpos.y));

	if(parent &&
	   newbbox.x > parent->GetValidArea().x &&
	   newbbox.y > parent->GetValidArea().y &&
	   newbbox.z < parent->GetValidArea().z &&
	   newbbox.w < parent->GetValidArea().w)
	    drag = true;

	if(drag)
	{
	    pos += Vectorf(mpos.x - prevmpos.x,
			   mpos.y - prevmpos.y);
	}
    }

    prevmpos = mpos;

    for(size_t i=0; i < children.size(); i++)
	children[i]->MouseMove(x-pos.x-theme->insideborder*3.0f, y-pos.y-theme->insideborder*3.0f);

    for(size_t i=0; i < children.size(); i++)
	if(children[i]->HasFocus())
	    ENDRETURN(true);

    if(Within() || dragging)
	focus = true;
    else
	focus = false;
    ENDRETURN(true);
}

bool GLWindow::Within()
{
    FUNC("GLWindow::Within");
    if(mpos.x > pos.x && mpos.x < (pos+dim).x &&
       mpos.y > pos.y+dim.y-theme->textsize.y*2.0f && mpos.y < (pos+dim).y &&
       draggable)
	ENDRETURN(true);
    ENDRETURN(false);
}

bool GLWindow::Render()
{
    FUNC("GLWindow::Render");
    glPushMatrix();

    glTranslatef(pos.x+dim.x/2, pos.y+dim.y/2, 0.0f);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_TEXTURE_2D);

    if(0 && Tex)
    {
	glEnable(GL_TEXTURE_2D);
	Tex->Bind();
    }
    else
	Tex = NULL;

    glLineWidth(theme->corner);
    float dimx = dim.x, dimy = dim.y;
    for(int i=0; i < 2; i++)
    {
	if(i == 1)
	{
	    if(focus && theme->highlight)
	    {
		glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.x,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.y,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.z,
			  theme->mainalpha);
	    }
	    else
	    {
		glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.x,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.y,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.z,
			  theme->mainalpha);
	    }
	    glDisable(GL_TEXTURE_2D);
	    glEnable(GL_BLEND);
	    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
	else
	{
	    if(focus && theme->highlight)
	    {
		glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.x,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.y,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.z,
			  theme->mainalpha);
	    }
	    else
	    {
		glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.x,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.y,
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.z,
			  theme->mainalpha);
	    }
	    if(Tex)
		glEnable(GL_TEXTURE_2D);
	    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}

	glBegin(GL_POLYGON);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(-dimx/2, -dimy/2);
	glTexCoord2f(0.0f, (1.0f-theme->corner));
	glVertex2f(-dimx/2, -dimy/2+dimy*(1.0f-theme->corner));
	glTexCoord2f(theme->corner, 1.0f);
	glVertex2f(-dimx/2+dimy*theme->corner, -dimy/2+dimy);
	glTexCoord2f(1.0f, 1.0f);
	glVertex2f(-dimx/2+dimx, -dimy/2+dimy);
	glTexCoord2f(1.0f, theme->corner);
	glVertex2f(-dimx/2+dimx, -dimy/2+dimy*theme->corner);
	glTexCoord2f(1.0f, 0.0f);
	glVertex2f(-dimx/2+dimx-dimy*theme->corner, -dimy/2);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(-dimx/2, -dimy/2);
	glEnd();
    }
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    glDisable(GL_TEXTURE_2D);

    glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.x,
	      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.y,
	      (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->bordercol.z,
	      theme->mainalpha);

    glBegin(GL_LINES);
    glVertex2f(-dimx/2, min(-dimy/2+dimy-theme->textsize.y*1.5f,
				   -dimy/2+dimy*(1.0f-theme->corner)));
    glVertex2f(-dimx/2+dimx, min(-dimy/2+dimy-theme->textsize.y*1.5f,
					-dimy/2+dimy*(1.0f-theme->corner)));
    glEnd();

    glColor3f(theme->textcol.x, theme->textcol.y, theme->textcol.z);
    glTranslatef(-dimx/2+dimy*theme->corner,
		 -dimy/2+dimy-theme->textsize.y*1.5f+theme->textsize.y/3.0f, 0.0f);
    font.Print(Text.c_str());

    glPopMatrix();

    glPushMatrix();
    glTranslatef(pos.x+theme->insideborder*3.0f, pos.y+theme->insideborder*3.0f, 0.0f);
    for(size_t i=0; i < children.size(); i++)
    {
	glPushMatrix();
	if(children[i]->GetDepth() > -0.5f)
	    children[i]->Render();
	glPopMatrix();
    }
    for(size_t i=0; i < children.size(); i++)
    {
	glPushMatrix();
	if(children[i]->GetDepth() < -0.5f)
	    children[i]->Render();
	glPopMatrix();
    }
    glPopMatrix();
    ENDRETURN(true);
}

bool GLWindow::MouseClick(bool left, bool middle, bool right)
{
    FUNC("GLWindow::MouseClick");
    int f = -1;
    for(size_t i=0; i < children.size(); i++)
	if(children[i]->MouseClick(left, middle, right) && children[i]->Unique())
	    f = i;
    if(f > -1)
    {
	for(size_t i=0; i < children.size(); i++)
	    if((signed)i != f)
		children[i]->SetFocussed(children[f]);
    }

    for(size_t i=0; i < children.size(); i++)
	if(children[i]->HasFocus())
	    ENDRETURN(false);

    if(left && focus && draggable)
	dragging = true;
    else
	dragging = false;

    ENDRETURN(true);
}

bool GLWindow::SetFocussed(GLWidget *widget)
{
    FUNC("GLWindow::SetFocussed");
    for(size_t i=0; i < children.size(); i++)
	children[i]->SetFocussed(widget);
    ENDRETURN(true);
}
