#include "glwidget.h"

bool Slider::Within()
{
    FUNC("Slider::Within");
    if(mpos.x > pos.x-theme->itemsize/2 && mpos.x < (pos+dim).x+theme->itemsize/2 &&
       mpos.y > pos.y-theme->itemsize/2 && mpos.y < (pos+dim).y+theme->itemsize/2)
	ENDRETURN(true);
    ENDRETURN(false);
}

bool Slider::MouseMove(float x, float y)
{
    FUNC("Slider::MouseMove");
    mpos = Vectorf(x, y);

    if(dragging && Within() && prevmpos.x > -5.0f && prevmpos.y > -5.0f)
    {
	// calculate the ratio of the width of the screen over the length of the slider
	float constnt;

	if(end.x == 0.0f)
	    constnt = (2.0f*float(viewport[2])/viewport[3])/(end.y * (2.0f*float(viewport[2])/viewport[3]));
	if(end.y == 0.0f)
	    constnt = (2.0f*float(viewport[2])/viewport[3])/(end.x * (2.0f*float(viewport[2])/viewport[3]));

	if(end.x == 0.0f && level + (mpos.y - prevmpos.y)*constnt > 0.0f
	   && level + (mpos.y - prevmpos.y)*constnt < 1.0f)
	    level += (mpos.y - prevmpos.y)*constnt;
	else if(end.y == 0.0f && level + (mpos.x - prevmpos.x)*constnt > 0.0f
		&& level + (mpos.x - prevmpos.x)*constnt < 1.0f)
	    level += (mpos.x - prevmpos.x)*constnt;
    }

    prevmpos = mpos;

    for(size_t i=0; i < children.size(); i++)
	children[i]->MouseMove(x, y);

    for(size_t i=0; i < children.size(); i++)
	if(children[i]->HasFocus())
	    ENDRETURN(true);

    if(Within() || dragging)
	focus = true;
    else
	focus = false;

    ENDRETURN(true);
}

bool Slider::MouseClick(bool left, bool middle, bool right)
{
    FUNC("Slider::MouseClick");
    int f = -1;
    for(size_t i=0; i < children.size(); i++)
	if(children[i]->MouseClick(left, middle, right) && children[i]->Unique())
	    f = i;
    if(f > -1)
    {
	for(size_t i=0; i < children.size(); i++)
	    if((signed)i != f)
		children[i]->SetFocussed(children[f]);
    }

    for(size_t i=0; i < children.size(); i++)
	if(children[i]->HasFocus())
	    ENDRETURN(false);

    if(focus && !lastleft)
	dragging = left;
    else
	focus = dragging = dragging && left;

    lastleft = left;
    lastmiddle = middle;
    lastright = right;
    ENDRETURN(true);
}

bool Slider::Render()
{
    FUNC("Slider::Render");
    assert(rendering);

    glPushMatrix();
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_DEPTH_TEST);

    if(focus && theme->highlight)
    {
	glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.x,
		  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.y,
		  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.z,
		  0.5f);
    }
    else
    {
	glColor4f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.x,
		  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.y,
		  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*theme->maincol.z,
		  0.5f);
    }

    glLineWidth(theme->border);

    glTranslatef(0.0f, theme->itemsize/4, 0.0f);

    if(end.x == 0.0f)
    {
	glBegin(GL_LINES);
	glVertex2f(pos.x, pos.y);
	glVertex2f(pos.x+end.x, pos.y+end.y+theme->itemsize/4.0f);
	glEnd();
    }
    else
    {
	glBegin(GL_LINES);
	glVertex2f(pos.x+theme->itemsize/2.0f, pos.y-theme->itemsize/4.0f);
	glVertex2f(pos.x+end.x+theme->itemsize/2.0f, pos.y+end.y-theme->itemsize/4.0f);
	glEnd();
    }

    theme->tex.Bind();
    glEnable(GL_TEXTURE_2D);

    if(end.x == 0.0f)
	glTranslatef(pos.x+level*end.x, pos.y+level*end.y, 0.0f);
    else
	glTranslatef(pos.x+level*end.x+theme->itemsize/2.0f, pos.y+level*end.y-theme->itemsize/4.0f, 0.0f);

    if(end.x == 0.0f)
	glRotatef(90.0f, 0.0f, 0.0f, 1.0f);

    glBegin(GL_QUADS);
    glTexCoord2f(slider.x, slider.y);
    glVertex3f(-theme->itemsize/2,-theme->itemsize/2,  0.0f);
    glTexCoord2f(slider.x, slider.w);
    glVertex3f(-theme->itemsize/2,  theme->itemsize/2, 0.0f);
    glTexCoord2f(slider.z, slider.w);
    glVertex3f( theme->itemsize/2,  theme->itemsize/2, 0.0f);
    glTexCoord2f(slider.z, slider.y);
    glVertex3f( theme->itemsize/2, -theme->itemsize/2, 0.0f);
    glEnd();

    glPopMatrix();

    ENDRETURN(true);
}
