#include "glwidget.h"

bool Textbox::Render()
{
    FUNC("Textbox::Render");
    assert(rendering);
    glPushMatrix();
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_DEPTH_TEST);
 
    glColor3f(theme->textcol.x, theme->textcol.y, theme->textcol.z);
    glTranslatef(pos.x+dim.x/2, pos.y+dim.y/2, 0.0f);
    glPushMatrix();

    float dimx = dim.x, dimy = dim.y;
    for(int i=0; i < 2; i++)
    {
	if(i == 1)
	{
	    if(focus && theme->highlight)
	    {
		glColor3f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.x),
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.y),
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.z));
	    }
	    else
	    {
		glColor3f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.x),
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.y),
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.z));
	    }
	    glDisable(GL_TEXTURE_2D);
	    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
	else
	{
	    if(focus && theme->highlight)
	    {
		glColor3f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.x),
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.y),
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.z));
	    }
	    else
	    {
		glColor3f((theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.x),
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.y),
			  (theme->brightfact*focusdt + theme->dimfact*(1.0f-focusdt))*(theme->textcol.z));
	    }
	    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
	glBegin(GL_POLYGON);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(-dimx/2, -dimy/2);

	glTexCoord2f(0.0f, (1.0f-theme->corner*3.0f));
	glVertex2f(-dimx/2, -dimy/2+dimy-theme->textsize.y*1.5f*theme->corner*3.0f);

	glTexCoord2f(theme->corner*3.0f, 1.0f);
	glVertex2f(-dimx/2+theme->textsize.y*1.5f*theme->corner*3.0f, -dimy/2+dimy);

	glTexCoord2f(1.0f, 1.0f);
	glVertex2f(-dimx/2+dimx, -dimy/2+dimy);

	glTexCoord2f(1.0f, theme->corner*3.0f);
	glVertex2f(-dimx/2+dimx, -dimy/2+theme->textsize.y*1.5f*theme->corner*3.0f);

	glTexCoord2f(1.0f, 0.0f);
	glVertex2f(-dimx/2+dimx-theme->textsize.y*1.5f*theme->corner*3.0f, -dimy/2);

	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(-dimx/2, -dimy/2);
	glEnd();
    }
    glPopMatrix();
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    glTranslatef(-dim.x/2+theme->textsize.x/2, -dim.y/2+theme->textsize.y/3.0f, 0.0f);
    glTranslatef(0.0f, theme->textsize.y*1.3f*(Height-1), 0.0f);
    glColor3f(1.0f-theme->textcol.x, 1.0f-theme->textcol.y, 1.0f-theme->textcol.z);
    font.Print(Text.c_str());
    if(start.find("\n") != string::npos)
	glTranslatef(font.Length(start.substr(start.rfind("\n")).c_str()),
		     -theme->textsize.y*1.3f*substr_count(start, "\n"), 0.0f);
    else
	glTranslatef(font.Length(start.c_str()), 0.0f, 0.0f);
    if(focus)
	font.Print("|");

    glPopMatrix();
    ENDRETURN(true);
}

bool Textbox::SetFocussed(GLWidget *widget)
{
    FUNC("Checkbox::SetFocussed");
    if(widget != this)
	focus = false;
    ENDRETURN(true);
}

bool Textbox::MouseMove(float x, float y)
{
    FUNC("Textbox::MouseMove");
    mpos = Vectorf(x, y);
    ENDRETURN(true);
}

bool Textbox::MouseClick(bool left, bool middle, bool right)
{
    FUNC("Textbox::MouseClick");
    if((Within() && left) || focus)
	focus = true;
    ENDRETURN(Within() && focus);
}

bool Textbox::TextKeyPress(bool keys[256])
{
    if(!focus)
	return true;

    if(keys[BACKSPACE] && start.size())
	start = start.substr(0, start.length()-1);

    if(keys[DELETEKEY] && end.size())
	end = end.substr(1);

    if(keys[LEFT] && start.length())
    {
	end = start[start.length()-1] + end;
	start = start.substr(0, start.length()-1);
    }
    if(keys[RIGHT] && end.length())
    {
	start += end[0];
	end = end.substr(1);
    }

    if(keys[HOME])
    {
	end = start+end;
	start = "";
    }

    if(keys[ENDKEY])
    {
	start = start+end;
	end = "";
    }

    if(keys[UP] && start.rfind("\n") != string::npos)
    {
	float distfromside = font.Length(start.substr(start.rfind("\n")).c_str());

	int linestart = start.substr(0, start.rfind("\n")).rfind("\n");

	if(linestart == (int)string::npos)
	    linestart = -1;

	string prevline = start.substr(linestart+1, start.rfind("\n")-linestart-1);

	int previdx = -3;
	float prevmin = 1000.0f;
	for(size_t i=0; i < prevline.size()+1; i++)
	{
	    float len = fabs(font.Length(prevline.substr(0, i).c_str())-distfromside);
	    if(len < prevmin)
		prevmin = len;
	    else
	    {
		previdx = i-1;
		break;
	    }
	}
	if(previdx == -3)
	    previdx = prevline.size();

	end = start.substr(linestart+1+previdx) + end;
	start = start.substr(0, linestart+1+previdx);
    }

    if(keys[DOWN] && end.find("\n") != string::npos)
    {
	float distfromside = 0.0f;

	if(start.rfind("\n") != string::npos)
	    distfromside = font.Length(start.substr(start.rfind("\n")).c_str());
	else
	    distfromside = font.Length(start.c_str());

	size_t lineend = end.substr(0, end.find("\n")).find("\n");

	if(lineend == string::npos)
	    lineend = end.size();

	string nextline = end.substr(end.find("\n")+1, lineend-end.find("\n"));

	int nextidx = -3;
	float nextmin = 1000.0f;
	for(size_t i=0; i < nextline.size()+1; i++)
	{
	    float len = fabs(font.Length(nextline.substr(0, i).c_str())-distfromside);
	    if(len < nextmin)
		nextmin = len;
	    else
	    {
		nextidx = i-1;
		break;
	    }
	}
	if(nextidx == -3)
	    nextidx = nextline.size();

	start = start + end.substr(0, end.find("\n")+1) + nextline.substr(0, nextidx);
	end = nextline.substr(nextidx);
    }

    if(keys[RETURN] && substr_count(Text, "\n")+1 < Height)
	start += "\n";

    string currentline;
    if(start.rfind("\n") != string::npos)
	currentline = start.substr(start.rfind("\n"));
    else
	currentline = start;

    if(font.Length(currentline.c_str())+font.Length("ww") > dim.x && 
       substr_count(Text, "\n")+1 < Height)
    {
	string prevline = start;
	if(start.rfind("\n") != string::npos)
	{
	    int linestart = start.rfind("\n");
	    
	    if(linestart == (int)string::npos)
		linestart = -1;
	    
	    prevline = start.substr(linestart+1, start.rfind("\n")-linestart-1);
	}

	string prevword = prevline;
	if(prevline.rfind(" ") != string::npos)
	    prevword = prevline.substr(prevline.rfind(" ")+1);

	if(prevword.size() < prevline.size())
	{
	    start = start.substr(0, start.length()-prevword.length()) + "\n" + prevword;

	    currentline = "";
	}
    }
    else if(font.Length(currentline.c_str())+font.Length("ww") <= dim.x)
    {
	for(int i=32; i < 126; i++)
	{
	    if(keys[i])
		start += char(i);
	}
    }
    Text = start + end;

    return true;
}
