#ifndef LOG_H
#define LOG_H

/*!
 * \file   log.h
 * \author  <baldurk@baldurk.org>
 * \date   Mon Jan  2 23:11:42 2006
 *
 * \brief  A logging class
 */

#include <string>
#include <fstream>
#include <cstdio>
#include <iostream>
using std::cerr;
using std::endl;

/*!
 * \brief A class that logs to a file or output
 * 
 * If DEBUG is defined, Logger provides a ostream-like
 * interface. One is defined and should always be used: Log.
 *
 * \todo figure out how to implement line and file information
 */

class Logger
{
  public:
    //! a default constructor
    Logger() {}
#if defined(DEBUG) || defined(LOG)
    static bool Load(std::string s)
    {
	if(s == "")
	    return false;
	else
	    freopen(s.c_str(), "w", stderr);
	return true;
    }

    ~Logger() { }
#else

    /*!
     * \brief A function to load the logger
     * 
     * If the parameter is blank, cerr is used, if it
     * is a string, that is the file used.
     */

    static bool Load(std::string s) { return true; }

    /*!
     * \brief A destructor
     * 
     * If a file has been opened, it is closed here.
     */

    ~Logger() { }
#endif
};

//! The default Logger for use.
#define Log cerr

#endif
