#include "model.h"

/*!
 * \file   plymodel.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:55:37 2006
 * 
 * \brief  Specific implementation of a model for PLY files
 */

int PLYModel::Load(string Filename)
{
    FUNC("PLYModel::Load");

    FILE *fp = fopen(Filename.c_str(), "r");

    if(!fp)
    {
	Log << "'" << Filename << "': Invalid model filename" << endl;;
	ENDRETURN(0);
    }

    if(GetLine(fp) != "ply")
    {
	Log << Filename << " is not a PLY file" << endl;
	ENDRETURN(1);
    }

    int NumVertices, NumNormals, NumTriangles;

    if(GetLine(fp) != "format ascii 1.0")
    {
	Log << "file is not in the supported format" << endl;
	ENDRETURN(1);
    }
    else
    {
	bool header = true;
	while(header)
	{
	    string line = GetLine(fp);
	    if(line.substr(0, 7) == "comment")
		continue;

	    if(line.substr(0, 7) == "element")
	    {
		if(line.substr(8, 6) == "vertex")
		{
		    istringstream iss;
		    iss.str(line.substr(15));
		    iss >> NumVertices;
		    NumNormals = NumVertices;
		    Normals.resize(NumNormals);
		    Vertices.resize(NumVertices);
		}
		if(line.substr(8, 4) == "face")
		{
		    istringstream iss;
		    iss.str(line.substr(13));
		    iss >> NumTriangles;
		    Triangles.resize(NumTriangles);
		}
	    }
	    if(line == "end_header")
		header = false;
	}

	istringstream iss;
	for(int i=0; i < NumVertices; i++)
	{
	    Log << 1 << ": " << float(i)/NumVertices << endl;
	    iss.str(GetLine(fp));
	    iss >> Vertices[i].pos.x >> Vertices[i].pos.y >>  Vertices[i].pos.z;
	}

	for(int i=0; i < NumTriangles; i++)
	{
	    Log << 2 << ": " << float(i)/NumTriangles << endl;
	    int dummy;
	    iss.str(GetLine(fp));
	    iss >> dummy
		>> Triangles[i].vertices[0]
		>> Triangles[i].vertices[1]
		>> Triangles[i].vertices[2];
	    Triangles[i].MaterialID = 0;
	    Triangles[i].smoothgroup = 0;
	}

	GenNormals();
    }

    NumMaterials = 0;

    ENDRETURN(0);
}

int PLYModel::Write(string f)
{
    assert("undefined write" && 0);
    return 0;
}
