#include "script.h"

ostream &operator <<(ostream &o, Variable var)
{
    FUNC("<< Variable");
    switch(var.type)
    {
	case VAR_INT:
	    ENDRETURN(o << var.i);
	case VAR_STRING:
	    ENDRETURN(o << "\"" << var.s << "\"");
	case VAR_FLOAT:
	    ENDRETURN(o << var.f);
	case VAR_ARRAY:
	    o << "Array {";
	    for(int i=0; i < (signed)var.v.size(); i++)
		o << " [" << i << "]=" << var.v[i];
	    o << " }";
	    ENDRETURN(o);
	case VAR_NULL:
	    ENDRETURN(o << "NULL");
	default:
	    assert(0 && "unhandled type");
    }
    ENDRETURN(o);
}

/*
  BNF for an expression

  [] = 0 or more
  <> = 0 or 1
  {} = one of this set

  line :: include | declaration | statement

  include :: #include "string"

  declaration :: funcdec | vardev

  funcdec :: type funcname([type string]) \n block
  funcname :: string

  vardec :: type varname;
  varname :: string

  statement :: expression | declaration | assignment | branch | block

  funccall :: string([expression]);

  assignment :: varname = expression;

  branch :: if(expression) statement |
            for(expression; expression; expression) statement |
	    while(expression) statement
	    do statement while(expression);

  block :: { statement [statement] }

  expression :: ( expression ) | funccall | varname | literal |
                expression op expression | expression op | op expression

  literal :: int | float | string
  varname :: string
  op :: {*,/,+,-,++,--,&&,||,&,|,^}

  Expression:

  expr
  1 + (5 + 4) * 10 + 4 / (4 + 4)
  expr op    expr
    1  +      (5 + 4) * 10 + 4 / (4 + 4)
               expr   op          expr
	       (5+4)  *            10 +         4 / (4 + 4)
               expr op expr        expr op      expr
	        5   +   4           10  +        4 /            (4 + 4)
		                                 expr op        expr
						   4   /        (4 + 4)
						                expr op expr
								  4   +   4

  1 + (5 + 4) * 10 + 4 / (4 + 4)
           =
  1 + 9 * 10 + 4 / 8
       =
  1 + 90 + 0.5
    =
  91.5

 */

bool Script::Start(string str, string test)
{
    return (str.find(test) == 0);
}

bool Script::Contains(string str, string test)
{
    return (str.find(test) != string::npos);
}

#define TrimFirst(string) string = string.substr(1)
#define TrimLast(string) string = string.substr(0,string.length()-1)

bool Script::StartNumber(string data)
{
    return Start(data, "0") || Start(data, "1") || Start(data, "2") ||
	Start(data, "3") || Start(data, "4") || Start(data, "5") || 
	Start(data, "6") || Start(data, "7") || Start(data, "8") || 
	Start(data, "9") || Start(data, ".") || Start(data, "-");
}

string Script::GetNumber(string &data)
{
    FUNC("Script::GetNumber");
    string ret = "";
    while(StartNumber(data))
    {
	ret += data[0];
	TrimFirst(data);
    }
    ENDRETURN(ret);
}

bool Script::IsNumber(string data)
{
    FUNC("Script::IsNumber");
    for(size_t i=0; i < data.length(); i++)
    {
	if(!StartNumber(data.substr(i)) && !StartSpace(data.substr(i)))
	    ENDRETURN(false);
    }
    ENDRETURN(data.length() > 0);
}

bool Script::StartSpace(string data)
{
    return Start(data, " ") || Start(data, "\t") || Start(data, "\n");
}

bool Script::IsSpace(string data)
{
    FUNC("Script::IsNumber");
    for(size_t i=0; i < data.length(); i++)
    {
	if(!StartSpace(data.substr(i)))
	    ENDRETURN(false);
    }
    ENDRETURN(true);
}

string Script::GetSpace(string &data)
{
    FUNC("Script::GetSpace");
    string ret = "";
    while(StartSpace(data))
    {
	ret += data[0];
	TrimFirst(data);
    }
    ENDRETURN(ret);
}

bool Script::StartOp(string data)
{
    return Start(data, "+") || Start(data, "-") || Start(data, "=")
	|| Start(data, "++") || Start(data, "-") || Start(data, "*")
	|| Start(data, "/") || Start(data, "^") || Start(data, "|")
	|| Start(data, "%") || Start(data, ">") || Start(data, "<")
	|| Start(data, ">=") || Start(data, "<=") || Start(data, "==")
	|| Start(data, "!=") || Start(data, "!") || Start(data, "&");
}

string Script::GetOp(string &data)
{
    FUNC("Script::GetOp");
    string ret = "";
    while(StartOp(data))
    {
	ret += data[0];
	TrimFirst(data);
    }
    ENDRETURN(ret);
}

bool Script::ContainsOp(string data)
{
    FUNC("Script::IsOp");
    for(size_t i=0; i < data.length(); i++)
    {
	if(StartOp(data.substr(i)))
	    ENDRETURN(true);
    }
    ENDRETURN(false);
}

bool Script::StartVariable(string data)
{
    return isalpha(data[0]) || Start(data, "-") || Start(data, "_");
}

string Script::GetVariableName(string &data)
{
    FUNC("Script::GetVariableName");
    string ret = "";
    while(StartVariable(data))
    {
	ret += data[0];
	TrimFirst(data);
    }
    ENDRETURN(ret);
}

string Script::GetVariable(string &data)
{
    FUNC("Script::GetVariable");
    string ret = "";
    while(StartVariable(data) || Start(data, "["))
    {
	while(data[0] == '[')
	{
	    ret += data[0];
	    TrimFirst(data);
	    ret += GetNumber(data);
	}
	ret += data[0];
	TrimFirst(data);
    }
    ENDRETURN(ret);
}

bool Script::IsVariable(string data)
{
    FUNC("Script::IsVariable");
    for(size_t i=0; i < data.length(); i++)
    {
	if(!StartVariable(data.substr(i)))
	    ENDRETURN(false);
    }
    ENDRETURN(data.length() > 0);
}

bool Script::StartString(string data)
{
    return Start(data, "\"");
}

string Script::GetString(string &data)
{
    FUNC("Script::GetString");
    string ret = "";
    if(StartString(data))
    {
	TrimFirst(data);
	while(data[0] != '"')
	{
	    if(data[0] == '\'')
	    {
		ret += data[1];
		TrimFirst(data);
		TrimFirst(data);
	    }
	    else
	    {
		ret += data[0];
		TrimFirst(data);
	    }
	}
	TrimFirst(data);
    }

    ENDRETURN(ret);
}

Variable *Script::GetVar(string varname)
{
    if(StartVariable(varname))
    {
	string work = varname;
	Variable *v = NULL;
	bool found = false;
	while(!IsSpace(work))
	{
	    if(StartVariable(work))
	    {
		v = &vars[GetVariableName(work)];
		found = true;
	    }
	    else if(Start(work, "["))
	    {
		if(!found)
		    break;
		TrimFirst(work);
		if(!StartNumber(work))
		    break;
		else
		{
		    string num = GetNumber(work);
		    istringstream iss;int i; iss.str(num); iss >> i;
		    v = &((*v)[i]);
		}
		while(StartSpace(work) || Start(work, "]"))
		    TrimFirst(work);
	    }
	    else
	    {
		found = false;
		return NULL;
	    }
	}
	if(found)
	{
	    return v;
	}
    }
    return NULL;
}

Variable Script::Eval(string expr)
{
    FUNC("Script::Eval");
    string expression = expr;

    GetSpace(expr);
    if(expr == "")
	ENDRETURN(Variable());
    
    while(Contains(expr, "("))
    {
	bool funccall = false;
	int funcstart = 0;
	string work = expr.substr(0,expr.find("("));
	string temp = "";

	while(!IsSpace(work) && work.length())
	{
	    if(StartNumber(work))
	    {
		temp = GetNumber(work);
		funcstart += temp.length();
	    }
	    else if(StartVariable(work))
	    {
		funccall = true;
		temp = GetVariable(work);
		funcstart += temp.length();
	    }
	    else if(StartString(work))
	    {
		temp = GetString(work);
		funcstart += temp.length();
	    }
	    else
	    {
		TrimFirst(work);
		funcstart++;
	    }
	}

	if(funccall)
	{
	    Variable funcval;

	    int bracketlevel = 1;
	    size_t i;
	    for(i=expr.find("(")+1; i < expr.length(); i++)
	    {
		if(bracketlevel == 0)
		    break;

		if(expr[i] == '(')
		    bracketlevel++;
		if(expr[i] == ')')
		    bracketlevel--;
	    }

	    string brackexp = expr.substr(expr.find("("), i-expr.find("("));
	    int bracklen = brackexp.length();

	    TrimFirst(brackexp);TrimLast(brackexp);
	    int numparams = 0, pos = 0;
	    pos = brackexp.find(",", pos);

	    while(pos != (signed)string::npos)
	    {
		numparams++;
		pos = brackexp.find(",", pos+1);
	    }

	    if(numparams > 0)
		numparams++;
	    else if(numparams == 0 && !IsSpace(brackexp))
		numparams++;

	    Variable *params = NULL;

	    if(numparams)
	    {
		params = new Variable[numparams];
		if(brackexp.find(",") == string::npos)
		    params[0] = Eval(brackexp);
		else
		    params[0] = Eval(brackexp.substr(0, brackexp.find(",")));

		string work = brackexp.substr(brackexp.find(",")+1);
		for(int i=1; i < numparams; i++)
		{
		    params[i] = Eval(work.substr(0, work.find(",")));
		    work = work.substr(work.find(",")+1);
		}
	    }

	    string funcname = expr.substr(funcstart - temp.length(), temp.length());

	    Function(func) = funcs[funcname];

	    if(!func)
		Log << "NULL exception: " << funcname << endl;
	    else
		funcval = func(funcname, numparams, params);

	    ostringstream oss;
	    oss << expr.substr(0, funcstart - temp.length())
		<< funcval
		<< expr.substr(expr.find("(")+bracklen);
	    expr = oss.str();
	}
	else
	{
	    string exp = expr;
	    int bracketlevel = 1;
	    size_t i;
	    for(i=expr.find("(")+1; i < expr.length(); i++)
	    {
		if(bracketlevel == 0)
		    break;

		if(expr[i] == '(')
		    bracketlevel++;
		if(expr[i] == ')')
		    bracketlevel--;
	    }

	    string brackexp = expr.substr(expr.find("("), i-expr.find("("));
	    TrimFirst(brackexp);TrimLast(brackexp);

	    Variable var = Eval(brackexp);
	    ostringstream oss;
	    oss << expr.substr(0, expr.find("("))
		<< var
		<< expr.substr(i, expr.length());
	    expr = oss.str();
	}
    }

    if(!ContainsOp(expr) || IsNumber(expr))
    {	
	istringstream iss;
	iss.str(expr);

	Variable var;
	float f;string s;
	iss >> f;
	if(iss.fail() == 0)
	{
	    if((int)f == f)
		var = (int)f;
	    else
		var = f;
	}
	else
	{
	    iss.clear();
	    iss.str(expr);
	    iss >> s;
	    var = Variable();
	    if(s[0] == '"' && s[s.length()-1] == '"')
	    {
		TrimFirst(s);TrimLast(s);
		var = s;
	    }
	    else
	    {
		var = *GetVar(s);
	    }
	}

	ENDRETURN(var);
    }

    vector<string> precedence;
    precedence.push_back("++");
    precedence.push_back("--");
    precedence.push_back("^");
    precedence.push_back("*");
    precedence.push_back("/");
    precedence.push_back("%");
    precedence.push_back("+");
    precedence.push_back("-");
    precedence.push_back("!");
    precedence.push_back("&&");
    precedence.push_back("||");
    precedence.push_back("<=");
    precedence.push_back("<");
    precedence.push_back(">=");
    precedence.push_back(">");
    precedence.push_back("==");
    precedence.push_back("!=");
    precedence.push_back("=");

    string work = expr;

    for(size_t i=0; i < precedence.size(); i++)
    {
	while(work.rfind(precedence[i]) != string::npos &&
	      !IsNumber(work))
	{
	    int pos = work.rfind(precedence[i]);
	    int startexp = 0, endexp = 0;
	    string exp2 = work.substr(pos+precedence[i].length());

	    endexp = pos+precedence[i].length();

	    while(StartSpace(exp2))
	    {
		endexp++;
		TrimFirst(exp2);
	    }

	    if(StartNumber(exp2))
	    {
		exp2 = GetNumber(exp2);
		endexp += exp2.length();
	    }
	    if(StartVariable(exp2))
	    {
		exp2 = GetVariable(exp2);
		endexp += exp2.length();
	    }
	    if(StartString(exp2))
	    {
		exp2 = string("\"") + GetString(exp2) + "\"";
		endexp += exp2.length();
	    }

	    string work2 = work.substr(0,pos);
	    string exp1 = "";

	    while(!IsSpace(work2) && work2.length())
	    {
		if(StartNumber(work2))
		{
		    exp1 = GetNumber(work2);
		    startexp += exp1.length();
		}
		else if(StartVariable(work2))
		{
		    exp1 = GetVariable(work2);
		    startexp += exp1.length();
		}
		else if(StartString(work2))
		{
		    exp1 = string("\"") + GetString(work2) + "\"";
		    startexp += exp1.length();
		}
		else
		{
		    TrimFirst(work2);
		    startexp++;
		}
	    }

	    startexp -= exp1.length();
	    string op = precedence[i];

	    Variable a = Eval(exp1);
	    Variable b = Eval(exp2);

	    if(a == Variable() && op != "=" && op != "++" && op != "--")
		Log << "NULL exception: " << exp1 << endl;

	    if(b == Variable() && op != "++" && op != "--")
		Log << "NULL exception: " << exp2 << endl;

	    Variable ret;

	    switch(op[0])
	    {
		case '+':
		{
		    if(op[1] == '+')
		    {
			ret = Eval(exp1 + " = " + exp1 + " + 1");
			break;
		    }
		    ret = a + b;break;
		}
		case '-':
		{
		    if(op[1] == '-')
		    {
			ret = Eval(exp1 + " = " + exp1 + " - 1");
			break;
		    }
		    ret = a - b;break;
		}
		case '*':
		    ret = a * b;break;
		case '/':
		    ret = a / b;break;
		case '^':
		    ret = a ^ b;break;
		case '%':
		    ret = a % b;break;
		case '=':
		{
		    if(op[1] == '=')
		    {
			ret = a == b;
			break;
		    }
		    Variable *one = GetVar(exp1);
		    if(one)
		    {
			ret = *one = b;
			break;
		    }
		    Log << "non l-value in assignment: " << exp1 << " = " << exp2 << endl;
		    break;
		}
		case '!':
		{
		    if(op[1] == '=')
		    {
			ret = a != b;
			break;
		    }
		    ret = !a;break;
		}
		case '<':
		{
		    if(op[1] == '=')
		    {
			ret = a <= b;
			break;
		    }
		    ret = a < b;break;
		}
		case '>':
		{
		    if(op[1] == '=')
		    {
			ret = a >= b;
			break;
		    }
		    ret = a > b;break;
		}
		case '|':
		{
		    if(op[1] == '|')
		    {
			ret = a || b;
			break;
		    }
		}
		case '&':
		{
		    if(op[1] == '&')
		    {
			ret = a && b;
			break;
		    }
		}
	    }

	    ostringstream oss;
	    oss << work.substr(0,startexp)
		<< ret
		<< work.substr(endexp);
	    work = oss.str();
	}
    }

    ENDRETURN(Eval(work));
}

bool Script::ParseTextFile(string file)
{
    FUNC("Script::Parse");

    ifstream ifs;
    ifs.open(file.c_str());

    string data;
    getline(ifs, data, '\0');

    while(data.size())
    {
	if(data.find(";") == string::npos)
	{
	    Log << "Error - missing ; in:" << endl << data.substr(0, data.find("\n")) << endl;
	    ENDRETURN(false);
	}
	string line = data.substr(0, data.find(";"));
	data = data.substr(data.find(";")+1, data.length());
	GetSpace(data);
	Variable ret = Eval(line);
	Log << line << " => " << ret << endl;
    }

    ifs.close();

#undef TrimFirst
#undef TrimLast

    ENDRETURN(true);
}

Script::Script()
{
    vars["false"] = 0;
    vars["true"] = 1;
    vars["pi"] = M_PI;
    vars["blah"][0] = 4;
    vars["blah"][1] = 6;
    vars["blah"][2][0] = 88;
    vars["blah"][2][1] = 89;
    vars["blah"][2][2] = 90;
    funcs["baz"] = &Script::baz;
}
