#include "texture.h"

bool TextureData::Load(int imwidth, int imheight, bool alpha, unsigned char *imdata)
{
    FUNC("TextureData::Load(imdata)");
    if(data)
	delete[] data;

    width = imwidth;
    height = imheight;

    data = new float[width*height*4];
    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	{
	    data[4*(x*height+y)+0] = float(imdata[(3+alpha)*(x*height+y)+0])/255.0f;
	    data[4*(x*height+y)+1] = float(imdata[(3+alpha)*(x*height+y)+1])/255.0f;
	    data[4*(x*height+y)+2] = float(imdata[(3+alpha)*(x*height+y)+2])/255.0f;
	    if(alpha)
		data[4*(x*height+y)+3] = float(imdata[(3+alpha)*(x*height+y)+3])/255.0f;
	    else
		data[4*(x*height+y)+3] = 1.0f;
	}
    ENDRETURN(true);
}

bool TextureData::Blank(int imwidth, int imheight)
{
    FUNC("TextureData::Blank");
    if(data)
	delete[] data;

    width = imwidth;
    height = imheight;

    data = new float[width*height*4];
    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	{
	    data[4*(x*height+y)+0] = 0.0f;
	    data[4*(x*height+y)+1] = 0.0f;
	    data[4*(x*height+y)+2] = 0.0f;
	    data[4*(x*height+y)+3] = 1.0f;
	}
    ENDRETURN(true);
}

bool TextureData::Invert(bool alpha)
{
    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	{
	    data[4*(x*height+y)+0] = 1.0f-data[4*(x*height+y)+0];
	    data[4*(x*height+y)+1] = 1.0f-data[4*(x*height+y)+1];
	    data[4*(x*height+y)+2] = 1.0f-data[4*(x*height+y)+2];
	    if(alpha)
		data[4*(x*height+y)+3] = 1.0f-data[4*(x*height+y)+3];
	}
}

float TextureData::Noise2D(int x, int y)
{
    int n = x + y * 50;
    n = (n<<13) ^ n;
    return (1.0f - ((n * (n * n * 15731 + 789221) + 1376312589) & 0x7fffffff) / 1073741824.0);
}

float TextureData::SmoothedNoise2D(int x, int y)
{
    return Noise2D(x, y)/4.0f +
	(Noise2D(x+1, y) + Noise2D(x-1, y) + Noise2D(x, y+1) + Noise2D(x, y-1))/8.0f +
	(Noise2D(x+1, y+1) + Noise2D(x+1, y-1) + Noise2D(x-1, y+1) + Noise2D(x-1, y-1))/16.0f;
}

float TextureData::Interpolate1D(float a, float b, float x)
{
    float f = (1.0f - cosf(x * M_PI)) * 0.5f;

    return a * (1.0f - f) + b * f;
}

float TextureData::InterpolatedNoise2D(float x, float y)
{
    float v1 = SmoothedNoise2D(int(x),   int(y));
    float v2 = SmoothedNoise2D(int(x)+1, int(y));
    float v3 = SmoothedNoise2D(int(x),   int(y)+1);
    float v4 = SmoothedNoise2D(int(x)+1, int(y)+1);

    float i1 = Interpolate1D(v1, v2, x-int(x));
    float i2 = Interpolate1D(v3, v4, x-int(x));

    return Interpolate1D(i1, i2, y-int(y));
}

bool TextureData::Perlin(int numoct, float Freq, float Amp, float freqmod, float ampmod)
{
    float freq = Freq, amp = Amp;

    for(int i=0; i < numoct; i++)
    {
	for(int x=0; x < width; x++)
	{
	    for(int y=0; y < height; y++)
	    {
		float X = amp*(InterpolatedNoise2D(float(x)/float(width*freq),
						   float(y)/float(height*freq)));
		data[4*(height*x+y)+0] += X;
		data[4*(height*x+y)+1] += X;
		data[4*(height*x+y)+2] += X;
	    }
	}
	freq *= freqmod;amp *= ampmod;
    }
}
