#include "terrain.h"

int Terrain::MidpointDisplace(int dim, float initmod, float mod,
			      float scalex, float scaley, float scalez)
{
    FUNC("Terrain::MidpointDisplace");

    dim++;

    vector<Vertex> Vertices;
    vector<Vectorf> Normals;
    vector<Triangle> Triangles;

    Vertices.resize(dim * dim);
    Normals.resize(dim * dim);

    for(int x=0; x < dim; x++)
	for(int y=0; y < dim; y++)
	{
	    Vertices[x*dim + y].pos =
		Vectorf(float(x)*float(dim/float(dim-1))-(dim/2.0f), -90000.0f,
			float(y)*float(dim/float(dim-1))-(dim/2.0f));

	    Normals[x*dim + y] = Vectorf(0.0f, 0.0f, 0.0f);

	    Vertices[x*dim + y].pos.x =
		Vertices[x*dim + y].pos.x * scalex / dim;
	    Vertices[x*dim + y].pos.y =
		Vertices[x*dim + y].pos.y * scaley / dim;
	    Vertices[x*dim + y].pos.z =
		Vertices[x*dim + y].pos.z * scalez / dim;
	}

    /*
      size = 5

      stride = 4
      O x x x O
      x x x x x
      x x * x x   diamond step
      x x x x x
      O x x x O

      stride = 4
      O x * x O
      x x x x x
      * x O x *   square step
      x x x x x
      O x * x O

      stride = 2
      O x O x O
      x * x * x
      O x O x O   diamond step
      x * x * x
      O x O x O

      stride = 2
      O * O * O
      * O * O *
      O * O * O   square step
      * O * O *
      O * O * O

      // order: 2,2 0,2 2,0 2,4 4,2 1,1 3,1 3,1 3,3 0,1 1,0 0,3 2,1 3,0

      Pseudocode:

      seed corners
      while(stride > 1)
      {
        diamond()
	square()
	halve stride
      }

      diamond()
      {
        for(int x=0; x < size-1; x += stride)
	{
          for(int y=0; y < size-1; y += stride)
	  {
	    verts[x+stride/2, y+stride/2] = average(verts[x, y],
	                                            verts[x+stride, y],
	                                            verts[x, y+stride],
	                                            verts[x+stride, y+stride])
	  }
	}
      }

      square()
      {
        for(int x=stride/2; x < size-1; x += stride)
	{
          for(int y=stride/2; y < size-1; y += stride)
	  {
	    wrap references, or clamp them
	    verts[x-stride/2, y] = average(verts[x, y],
	                                   verts[x-stride/2, y-stride/2],
					   verts[x-stride/2, y+stride/2],
					   verts[x-stride, y])
	    verts[x, y-stride/2] = average(verts[x+stride/2, y-stride/2],
	                                   verts[x, y-stride],
					   verts[x, y],
					   verts[x-stride/2, y-stride/2])
	  }
	}
      }

     */

    float amplitude = scaley*initmod;
    int stride = dim-1;

#define COORD(x, y) ((x>=0?(x<dim?x:x-dim+1):x+dim-1)*dim+(y>=0?(y<dim?y:y-dim+1):y+dim-1))

    Vertices[COORD(0, 0)].pos.y = 
    Vertices[COORD(dim-1, 0)].pos.y = 
    Vertices[COORD(0, dim-1)].pos.y =
    Vertices[COORD(dim-1, dim-1)].pos.y = 0.0f;

    amplitude *= mod;

    while(stride > 1)
    {
	for(int x=0; x < dim-1; x += stride)
	{
	    for(int y=0; y < dim-1; y += stride)
	    {
		Vertices[COORD(x+stride/2, y+stride/2)].pos.y =
		    (Vertices[COORD(x, y)].pos.y +
		     Vertices[COORD(x+stride, y)].pos.y +
		     Vertices[COORD(x, y+stride)].pos.y +
		     Vertices[COORD(x+stride, y+stride)].pos.y) / 4.0f;

		Vertices[COORD(x+stride/2, y+stride/2)].pos.y +=
		    RAND(-amplitude, amplitude);
	    }
	}

	for(int x=stride/2; x < dim-1; x += stride)
	{
	    for(int y=stride/2; y < dim-1; y += stride)
	    {
		Vertices[COORD(x-stride/2, y)].pos.y =
		    (Vertices[COORD(x, y)].pos.y +
		     Vertices[COORD(x-stride/2, y-stride/2)].pos.y +
		     Vertices[COORD(x-stride/2, y+stride/2)].pos.y +
		     Vertices[COORD(x-stride, y)].pos.y) / 4.0f;

		Vertices[COORD(x, y-stride/2)].pos.y =
		    (Vertices[COORD(x+stride/2, y-stride/2)].pos.y +
		     Vertices[COORD(x, y-stride)].pos.y +
		     Vertices[COORD(x, y)].pos.y +
		     Vertices[COORD(x-stride/2, y-stride/2)].pos.y) / 4.0f;

		Vertices[COORD(x+stride/2, y)].pos.y =
		    (Vertices[COORD(x, y)].pos.y +
		     Vertices[COORD(x+stride/2, y-stride/2)].pos.y +
		     Vertices[COORD(x+stride/2, y+stride/2)].pos.y +
		     Vertices[COORD(x+stride, y)].pos.y) / 4.0f;

		Vertices[COORD(x, y+stride/2)].pos.y =
		    (Vertices[COORD(x+stride/2, y+stride/2)].pos.y +
		     Vertices[COORD(x, y+stride)].pos.y +
		     Vertices[COORD(x, y)].pos.y +
		     Vertices[COORD(x-stride/2, y+stride/2)].pos.y) / 4.0f;

		Vertices[COORD(x-stride/2, y)].pos.y +=
		    RAND(-amplitude, amplitude);

		Vertices[COORD(x, y-stride/2)].pos.y +=
		    RAND(-amplitude, amplitude);

		Vertices[COORD(x+stride/2, y)].pos.y +=
		    RAND(-amplitude, amplitude);

		Vertices[COORD(x, y+stride/2)].pos.y +=
		    RAND(-amplitude, amplitude);
	    }
	}

	for(int x=stride/2; x < dim-1; x += stride)
	{
	    Vertices[COORD(x, dim-1)].pos.y = 
		Vertices[COORD(x, 0)].pos.y;
	}

	for(int y=stride/2; y < dim-1; y += stride)
	{
	    Vertices[COORD(dim-1, y)].pos.y = 
		Vertices[COORD(0, y)].pos.y;
	}

	stride /= 2;
	amplitude *= mod;
    }

    for(int x=0; x < dim-1; x++)
	for(int y=0; y < dim-1; y++)
	{
	    Triangle tri1, tri2;
	    tri1.vertices[2] =     x*dim + y;
	    tri1.vertices[1] = (x+1)*dim + y;
	    tri1.vertices[0] =     x*dim + (y+1);
	    tri1.normals[2] =     x*dim + y;
	    tri1.normals[1] = (x+1)*dim + y;
	    tri1.normals[0] =     x*dim + (y+1);

	    tri2.vertices[2] = (x+1)*dim + y;
	    tri2.vertices[1] = (x+1)*dim + (y+1);
	    tri2.vertices[0] =     x*dim + (y+1);
	    tri2.normals[2] = (x+1)*dim + y;
	    tri2.normals[1] =     x*dim + (y+1);
	    tri2.normals[0] = (x+1)*dim + (y+1);

	    Triangles.push_back(tri1);
	    Triangles.push_back(tri2);
	}

    assert(0);

    //model.Load(mesh, 1, NULL, 0);

    //model.GenNormals();

    //model.SetMode(IMMEDIATE);
    //model.SetMode(SEQUENTIAL);

    ENDRETURN(0);
}

int Terrain::Heightmap(string filename, float scalex, float scaley, float scalez)
{
    FUNC("Terrain::Heightmap");

    TextureData data;
    data.Load(filename.c_str());

    data.Blur(4);

    vector<Vertex> Vertices;
    vector<Vectorf> Normals;
    vector<Triangle> Triangles;

    Vertices.resize(data.Width() * data.Height());
    Normals.resize(data.Width() * data.Height());

    for(int x=0; x < data.Width(); x++)
	for(int y=0; y < data.Height(); y++)
	{
	    Vertices[x*data.Width() + y].pos =
		Vectorf(x-(data.Width()/2.0f), 0.0f,
			y-(data.Height()/2.0f));

	    Vertices[x*data.Width() + y].pos.y =
		(data.Width()/2.0f)*data.Data()[4*(x*data.Width()+y)];
	    Normals[x*data.Width() + y] = Vectorf(0.0f, 1.0f, 0.0f);

	    Vertices[x*data.Width() + y].pos.x =
		Vertices[x*data.Width() + y].pos.x * scalex / data.Width();
	    Vertices[x*data.Width() + y].pos.y =
		Vertices[x*data.Width() + y].pos.y * scaley / data.Width();
	    Vertices[x*data.Width() + y].pos.z =
		Vertices[x*data.Width() + y].pos.z * scalez / data.Height();
	}

    /*
      1--------2
      |       /|
      |      / |
      |     /  |           1, 2, 3
      |    /   |              &
      |   /    |           2, 4, 3
      |  /     |
      | /      |
      |/       |
      3--------4
    */

    for(int x=0; x < data.Width()-1; x++)
	for(int y=0; y < data.Height()-1; y++)
	{
	    Triangle tri1, tri2;
	    tri1.vertices[2] =     x*data.Width() + y;
	    tri1.vertices[1] = (x+1)*data.Width() + y;
	    tri1.vertices[0] =     x*data.Width() + (y+1);
	    tri1.normals[2] =     x*data.Width() + y;
	    tri1.normals[1] = (x+1)*data.Width() + y;
	    tri1.normals[0] =     x*data.Width() + (y+1);

	    tri2.vertices[2] = (x+1)*data.Width() + y;
	    tri2.vertices[1] = (x+1)*data.Width() + (y+1);
	    tri2.vertices[0] =     x*data.Width() + (y+1);
	    tri2.normals[2] = (x+1)*data.Width() + y;
	    tri2.normals[1] =     x*data.Width() + (y+1);
	    tri2.normals[0] = (x+1)*data.Width() + (y+1);

	    Triangles.push_back(tri1);
	    Triangles.push_back(tri2);
	}

    unsigned char tmpdata[] = {0, 0, 0, 0,    0, 0, 0, 0,
			       0, 0, 0, 0,    0, 0, 0, 0 };

    assert(0);

    data.Load(2, 2, true, tmpdata);

    //model.Load(mesh, 1, NULL, 0);
    //GenNormals();

    //model.SetMode(IMMEDIATE);
    //model.SetMode(SEQUENTIAL);

    ENDRETURN(0);
}

int Terrain::PerlinNoise(int dim, int numoct, float freq, float amp, float freqmod,
			 float ampmod, float scalex, float scaley, float scalez)
{
    FUNC("Terrain::Heightmap");

    TextureData data;
    data.Blank(dim, dim);
    data.Perlin(numoct, freq, amp, freqmod, ampmod);

    vector<Vertex> Vertices;
    vector<Vectorf> Normals;
    vector<Triangle> Triangles;

    Vertices.resize(data.Width() * data.Height());
    Normals.resize(data.Width() * data.Height());

    for(int x=0; x < data.Width(); x++)
	for(int y=0; y < data.Height(); y++)
	{
	    Vertices[x*data.Width() + y].pos =
		Vectorf(x-(data.Width()/2.0f), 0.0f,
			y-(data.Height()/2.0f));

	    Vertices[x*data.Width() + y].pos.y =
		(data.Width()/2.0f)*data.Data()[4*(x*data.Width()+y)];
	    Normals[x*data.Width() + y] = Vectorf(0.0f, 1.0f, 0.0f);

	    Vertices[x*data.Width() + y].pos.x =
		Vertices[x*data.Width() + y].pos.x * scalex / data.Width();
	    Vertices[x*data.Width() + y].pos.y =
		Vertices[x*data.Width() + y].pos.y * scaley / data.Width();
	    Vertices[x*data.Width() + y].pos.z =
		Vertices[x*data.Width() + y].pos.z * scalez / data.Height();
	}

    /*
      1--------2
      |       /|
      |      / |
      |     /  |           1, 2, 3
      |    /   |              &
      |   /    |           2, 4, 3
      |  /     |
      | /      |
      |/       |
      3--------4
    */

    for(int x=0; x < data.Width()-1; x++)
	for(int y=0; y < data.Height()-1; y++)
	{
	    Triangle tri1, tri2;
	    tri1.vertices[2] =     x*data.Width() + y;
	    tri1.vertices[1] = (x+1)*data.Width() + y;
	    tri1.vertices[0] =     x*data.Width() + (y+1);
	    tri1.normals[2] =     x*data.Width() + y;
	    tri1.normals[1] = (x+1)*data.Width() + y;
	    tri1.normals[0] =     x*data.Width() + (y+1);

	    tri2.vertices[2] = (x+1)*data.Width() + y;
	    tri2.vertices[1] = (x+1)*data.Width() + (y+1);
	    tri2.vertices[0] =     x*data.Width() + (y+1);
	    tri2.normals[2] = (x+1)*data.Width() + y;
	    tri2.normals[1] =     x*data.Width() + (y+1);
	    tri2.normals[0] = (x+1)*data.Width() + (y+1);

	    Triangles.push_back(tri1);
	    Triangles.push_back(tri2);
	}

    unsigned char tmpdata[] = {0, 0, 0, 0,    0, 0, 0, 0,
			       0, 0, 0, 0,    0, 0, 0, 0 };

    assert(0);

    data.Load(2, 2, true, tmpdata);

    //model.Load(mesh, 1, NULL, 0);
    //model.GenNormals();

    //model.SetMode(IMMEDIATE);
    //model.SetMode(SEQUENTIAL);

    ENDRETURN(0);
}

int Terrain::Render()
{
    FUNC("Terrain::Render");
    model.Render();
    ENDRETURN(0);
}
