#ifndef TERRAIN_H
#define TERRAIN_H

#include "general.h"
#include "texture.h"
#include "model.h"

/*!
 * \file   terrain.h
 * \author  <baldurk@baldurk.org>
 * \date   Tue Feb  7 15:08:37 2006
 * 
 * \brief  A class to generate terrain
 */

//! A class to generate and render terrain
class Terrain
{
  public:
    /*!
     * Generates the terrain from a texture. Optionally scaling the
     * bounding box from (1, 0.5, 1)
     * 
     * \param filename the name of the texture
     * \param scalex   how to scale the terrain in the x direction
     * \param scaley   how to scale the terrain in the y direction
     * \param scalez   how to scale the terrain in the z direction
     */
    int Heightmap(string filename,
		  float scalex = 1.0f,
		  float scaley = 1.0f,
		  float scalez = 1.0f);

    /*!
     * Generates the terrain from perlin noise. Optionally scaling the
     * bounding box from (1, 0.5, 1)
     * 
     * \param dim     the dimensions of the terrain
     * \param numoct  the dimensions of the terrain
     * \param freq    the dimensions of the terrain
     * \param amp     the dimensions of the terrain
     * \param freqmod the dimensions of the terrain
     * \param ampmod  the dimensions of the terrain
     * \param scalex  how to scale the terrain in the x direction
     * \param scaley  how to scale the terrain in the y direction
     * \param scalez  how to scale the terrain in the z direction
     */
    int PerlinNoise(int dim,
		    int numoct,
		    float freq,
		    float amp,
		    float freqmod,
		    float ampmod,
		    float scalex = 1.0f,
		    float scaley = 1.0f,
		    float scalez = 1.0f);

    /*!
     * Generates the terrain via midpoint displacement
     * 
     * \param dim     the dimensions of the terrain
     * \param initmod the initial modifier for the amplitude (multiplied by scaley)
     * \param mod     the modifier applied 
     * \param scalex  how to scale the terrain in the x direction
     * \param scaley  how to scale the terrain in the y direction
     * \param scalez  how to scale the terrain in the z direction
     */
    int MidpointDisplace(int dim,
			 float initmod,
			 float mod,
			 float scalex = 1.0f,
			 float scaley = 1.0f,
			 float scalez = 1.0f);

    //! Render the terrain
    int Render();
  private:
    //! the model data for the terrain
    Model model;
};

#endif
