#include "texture.h"

/*!
 * \file   texture.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:56:38 2006
 * 
 * \brief  details of a Texture, between the data and the API
 */

//map<TextureDetails, GLuint> Texture::textures;
vector<Tex> Texture::textures;
vector<short> Texture::refcounts;

bool Texture::Load(TextureDetails details)
{
    FUNC("Texture::Load(details)");

    for(int i=0; i < (signed)textures.size(); i++)
    {
	if(textures[i].details == details)
	{
	    if(ID == textures[i].ID)
	    {
		Log << "reassigning this texture" << endl;
		ENDRETURN(true);
	    }

	    if(glIsTexture(ID))
	    {
		while(ID >= refcounts.size())
		    refcounts.push_back(0);
		refcounts[ID]--;
	    }
	    ID = textures[i].ID;
	    if(glIsTexture(ID))
	    {
		while(ID >= refcounts.size())
		    refcounts.push_back(0);
		refcounts[ID]++;
	    }
	    Log << "not reloading texture" << endl;
	    ENDRETURN(true);
	}
    }

    ID = 0;

    TextureData data;

    data.Load(details);

    Load(data, details);

    bool found = false;
    for(int i=0; i < (signed)textures.size(); i++)
	if(textures[i].details == details)
	{
	    textures[i].ID = ID;
	    found = true;
	}

    Tex t;t.details = details;t.ID = ID;

    if(!found)
	textures.push_back(t);

    while(ID >= refcounts.size())
	refcounts.push_back(0);
    refcounts[ID]++;

    ENDRETURN(true);
}

bool TextureData::Load(TextureDetails details)
{
    FUNC("TextureData::Load(details)");
    if(details.name.find("tga") != string::npos)
    {
	ENDRETURN(LoadTGA(details));
    }
    else if(details.name.find("png") != string::npos)
    {
	ENDRETURN(LoadPNG(details));
    }
    else
    {
	Log << details.name << ": Unsupported texture format" << endl;
	ENDRETURN(false);
    }
    ENDRETURN(false);
}

bool TextureData::LoadTGA(TextureDetails details)
{
    FUNC("TextureData::LoadTGA");
    FILE *fp;

    if(!(fp = fopen(details.name.c_str(), "rb")))
    {
	Log << "Error loading texture tga file '" << details.name << "'" << endl;
	ENDRETURN(false);
    }

    unsigned char dummy;
	
    for(int i=0; i < 3; i++)
	fread(&dummy, sizeof(unsigned char), 1, fp);

    //! check the type, the 3rd byte read
    if(dummy != 10)
    {
	Log << "Please use compressed TGAs" << endl;
	ENDRETURN(false);
    }

    for(int i=0; i < 9; i++)
	fread(&dummy, sizeof(unsigned char), 1, fp);

    unsigned char header[6];

    fread(header, sizeof(unsigned char), 6, fp);

    width = (header[1] << 8) + header[0];
    height = (header[3] << 8) + header[2];
    int bpp = header[4];

    unsigned char *tgadata = new unsigned char[width * height * (bpp/8)];
    int pixel = 0, ref = 0, numPix = width * height;
    unsigned char *temp = new unsigned char[bpp];
    while(pixel < numPix)
    {
	unsigned char chunk;
	fread(&chunk, sizeof(unsigned char), 1, fp);
	if(chunk < 128)
	{
	    chunk++;
	    for(short i = 0; i < chunk; i++)
	    {
		fread(temp, sizeof(unsigned char), (bpp/8), fp);
		tgadata[ref]   = temp[2];
		tgadata[ref+1] = temp[1];
		tgadata[ref+2] = temp[0];
		tgadata[ref+3] = temp[3];

		ref += (bpp/8);
		pixel++;
	    }
	}
	else
	{
	    chunk -= 127;
	    fread(temp, sizeof(unsigned char), (bpp/8), fp);
			
	    for(short i = 0; i < chunk; i++)
	    {
		tgadata[ref]   = temp[2];
		tgadata[ref+1] = temp[1];
		tgadata[ref+2] = temp[0];
		tgadata[ref+3] = temp[3];
	
		ref += (bpp/8);
		pixel++;
	    }
	}
    }
    delete[] temp;
    fclose(fp);

    delete [] data;
    data = new float[width*height*4];
    for(int x=0; x < width; x++)
    {
	for(int y=0; y < height; y++)
	{
	    if(bpp == 24)
	    {
		data[(x*height+y)*4+0] = tgadata[(x*height+y)*3+0]/255.0f;
		data[(x*height+y)*4+1] = tgadata[(x*height+y)*3+1]/255.0f;
		data[(x*height+y)*4+2] = tgadata[(x*height+y)*3+2]/255.0f;
		data[(x*height+y)*4+3] = 1.0f;
	    }
	    else if(bpp == 32)
	    {
		data[(x*height+y)*4+0] = tgadata[(x*height+y)*4+0]/255.0f;
		data[(x*height+y)*4+1] = tgadata[(x*height+y)*4+1]/255.0f;
		data[(x*height+y)*4+2] = tgadata[(x*height+y)*4+2]/255.0f;
		data[(x*height+y)*4+3] = tgadata[(x*height+y)*4+3]/255.0f;
	    }
	}
    }

    ENDRETURN(true);
}

unsigned char PNGImage::GetByte()
{
#ifdef DEBUG_SLOW
    FUNC("PNGImage::GetByte");
#endif
    if(byte == 0)
    {
	byte++;
#ifdef DEBUG_SLOW
	ENDRETURN((data[cnt] & 0xff00) >> 0x8);
#else
	return (data[cnt] & 0xff00) >> 0x8;
#endif
    }

    byte = 0;cnt++;
#ifdef DEBUG_SLOW
    ENDRETURN(data[cnt-1] & 0x00ff);
#else
    return data[cnt-1] & 0x00ff;
#endif
}

unsigned int PNGImage::GetLong()
{
#ifdef DEBUG_SLOW
    FUNC("PNGImage::GetLong");
#endif
    unsigned char a, b, c, d;
    a = GetByte();b = GetByte();c = GetByte();d = GetByte();
#ifdef DEBUG_SLOW
    ENDRETURN((unsigned int)(a << 0x18 | b << 0x10 | c << 0x8 | d));
#else
    return (unsigned int)(a << 0x18 | b << 0x10 | c << 0x8 | d);
#endif
}

bool TextureData::LoadPNG(TextureDetails details)
{
    FUNC("TextureData::LoadPNG");

    FILE *fp = fopen(details.name.c_str(), "rb");

    if(!fp)
    {
	Log << "Error opening " << details.name << endl;
	ENDRETURN(false);
    }

    fseek(fp, 0, SEEK_END);
    long fsize = (long)ftell(fp)/sizeof(unsigned short);
    fseek(fp, 0, SEEK_SET);

    PNGImage PNG;

    PNG.data = new unsigned short[fsize+2];

    // read the file into the PNG.data buffer
    unsigned short s;
    int cnt = 0;
    for(int i=0; i < fsize;i+=5)
    {
	if(!feof(fp))
	{
	    fread(&s, sizeof(s), 1, fp);
	    PNG.data[cnt++] = ntohs(s);
	}
	if(!feof(fp))
	{
	    fread(&s, sizeof(s), 1, fp);
	    PNG.data[cnt++] = ntohs(s);
	}
	if(!feof(fp))
	{
	    fread(&s, sizeof(s), 1, fp);
	    PNG.data[cnt++] = ntohs(s);
	}
	if(!feof(fp))
	{
	    fread(&s, sizeof(s), 1, fp);
	    PNG.data[cnt++] = ntohs(s);
	}
	if(!feof(fp))
	{
	    fread(&s, sizeof(s), 1, fp);
	    PNG.data[cnt++] = ntohs(s);
	}
    }

    fclose(fp);

    unsigned int id[2];
    id[0] = PNG.GetLong();id[1] = PNG.GetLong();

    // check the 8 byte PNG header is correct
    if(id[0] != 0x89504e47 ||
       id[1] != 0x0D0A1A0A)
    {
	Log << details.name << " is not a valid png file - header invalid" << endl;
	ENDRETURN(false);
    }

    width=1, height=1;
    unsigned char *streamdata=new unsigned char[fsize*sizeof(unsigned short)];
    unsigned long streamlen=0, streamcurrent=0;

    while(PNG.cnt < (signed)fsize)
    {
	PNGChunk c;
	c.len = PNG.GetLong();
        c.type = PNG.GetLong();
	if(c.len)
	    c.data = new unsigned char[c.len];

	for(int j=0; j < c.len; j++)
	    c.data[j] = PNG.GetByte();
	c.crc = PNG.GetLong();

	//! \todo check crc here

	// IHDR = I(0x49) H(0x48) D(0x44) R(0x52)
	if(c.type == 0x49484452)
	{
	    // get width
	    width = long(c.data[0] << 0x18 | c.data[1] << 0x10 | c.data[2] << 0x8 | c.data[3]);
	    // get height
	    height = long(c.data[4] << 0x18 | c.data[5] << 0x10 | c.data[6] << 0x8 | c.data[7]);

	    // check bitdepth is 6 (RGBA truecolour)
	    // check compression is method 0 (zlib)
	    // check filter method is 0 (normal filtering)
	    // check interlace method is 0 (no interlacing)
	    if(c.data[9] != 6 ||
	       c.data[10] != 0 ||
	       c.data[11] != 0 ||
	       c.data[12] != 0)
	    {
		delete[] c.data;
		Log << "Unsupported format of PNG. Make sure it is RGBA, non-interlaced." << endl;
		ENDRETURN(false);
	    }
	}

	// IDAT = a new chunk will be put into streamdata,
	//   old chunks concatenated on the end
	if(c.type == 0x49444154)
	{
	    // append IDAT bytes into buffer
	    memcpy(&streamdata[streamcurrent], c.data, c.len);
	    // update size of stream
	    streamlen += c.len;
	    streamcurrent += c.len;
	}

	if(c.len)
	    delete[] c.data;
    }

    // the destination (ie. uncompressed) data buffer and length
    unsigned char *dest;
    unsigned long destlen = 32;

    // check zlib header
    if(streamdata[0] & 0xf != 8 || (streamdata[0] & 0xf0) >> 4 != 7)
    {
	Log << "Invalid compression format." << endl;
	ENDRETURN(false);
    }

    int ret = 0;

    dest = new unsigned char[(width*height*4)+width];

    // do the actual decompression
    ret = Inflate(dest, &destlen, &streamdata[2], streamlen);

    //ret = uncompress(dest, (unsigned long *)&destlen,
    // &streamdata[2], streamlen);
    if(ret != 0)
    {
	Log << "Could not decompress PNG data. Error code: " << ret << endl;
	ENDRETURN(false);
    }

    // current byte
    int cb = 0;
    // output byte
    int ob = 0;

    // the final image data to be passed to opengl
    unsigned char *image = new unsigned char[width*height*4];

    // perform filtering row by row.
    for(int i=0; i < height; i++)
    {
	// if the first byte in the row is 0 there is no filtering
	// skip this row - perform no filtering on it
	if(dest[cb] == 0)
	{
	    cb++; // skip filter byte
	    memcpy(&image[ob], &dest[cb], width*4);
	    cb += width*4;
	    ob += width*4;
	}
	// if the first byte is 1 then this row is filtered by
	// the "Sub" method. the final byte is the current byte
	// plus the matching byte in the previous pixel.
	// the leftmost pixel is presumed to have a previous pixel
	// of (0, 0, 0, 0). The final value is moded to 256 to
	// keep it in the byte range
	else if(dest[cb] == 1)
	{
	    cb++; // skip filter byte
	    memcpy(&image[ob], &dest[cb], 4);
	    cb += 4; // skip first pixel
	    ob += 4;

	    // for every byte
	    for(int j=4; j < width*4; j++)
	    {
		// add the previous pixel's byte to this one
		// e.g. if this is a red byte, add the last
		// red byte to this pixel.
		// finally divide by 256 and take the remainder
		// so that the value stays in the 0 - 255 range
		image[ob++] = (dest[cb] + image[ob-4])%256;
		cb++;
	    }
	}
	// if the first byte is 1 then this row is filtered
	// by the "Up" method. This is the same as the "Sub" method
	// but refers to the pixel above, not to the left
	else if(dest[cb] == 2)
	{
	    cb++; // skip filter byte
	    if(i > 0) // check this is not the first row
		for(int j=0; j < width*4; j++)
		{
		    // width*4+1 is the # bytes in a row
		    // so add the above byte to this one and mod with 256
		    image[ob++] = (dest[cb] + image[ob-(width*4)])%256;
		    cb++;
		}
	}
	// if the first byte is 3 then use the "Average" filter
	// this filter takes the average of the left and up bytes
	// in the same way as above. Pixels/bytes outside the image
	// are assumed to be 0
	else if(dest[cb] == 3)
	{
	    cb++; // skip filter byte

	    // for every byte in this row
	    for(int j=0; j < width*4; j++)
	    {
		float foo = 0.0f;
		// the if statements just check which of the up or
		// left bytes will be outside the image, if any,
		// and perform the same calculation but with
		// 0 replaced. The first if uses the image for both
		// bytes. It simply adds both up and left then divides
		// by 2.


		if(j > 3)// && i > 0)
		    foo = dest[cb] + 
			((image[ob-4] + image[ob-(width*4)])/2);
		else if(j <= 3)// && i > 0)
		    foo = dest[cb] + 
			((0 + image[ob-(width*4)])/2);
		/*else if(j > 3 && i == 0)
		    foo = dest[cb] + 
			((image[ob-4] + 0)/2);
		else if(j <= 3 && i == 0)
		    foo = dest[cb] + 
		    ((0 + 0)/2);*/
		    
		// after the above, round down to the nearest int, then
		// mod by 256 again
		image[ob++] = (unsigned char)(foo) % 256;
		cb++;
	    }
	}
	// 4 is the paeth filter and is more complex.
	else if(dest[cb] == 4)
	{
	    cb++; // skip filter byte

	    for(int j=0; j < width*4; j++)
	    {
		// a is the byte to the left, b above, c is left and above
		// first we get them
		unsigned char a=0, b=0, c=0;
		int pa, pb, pc, p;
		if(i > 0)
		{
		    if(j > 3)
			c = image[ob-(width*4)-4];
		    else
			c = 0;

		    b = image[ob-(width*4)];
		}
		else
		    b = c = 0;

		if(j > 3)
		    a = image[ob-4];
		else
		    a = 0;

		// then work out some intermediate values
		p = b - c;
		pc = a - c;

		// get the absolute values of the distances
		pa = p < 0 ? -p : p;
		pb = pc < 0 ? -pc : pc;
		pc = (p + pc) < 0 ? -(p + pc) : p + pc;

		// if a is smallest, take it. If not, check b. Finally check c.
		// this order is important

		p = (pa <= pb && pa <=pc) ? a : (pb <= pc) ? b : c;

		// add the stored byte to the p value worked out,
		// and take the low-end byte.

		image[ob] = (unsigned char)(((int)(dest[cb]) + p) & 0xff);

		// move to the next byte
		ob++;
		cb++;
	    }
	}
	// if the first byte isn't 0 - 4 something went wrong
	else
	{
	    delete[] dest;
	    delete[] streamdata;
	    Log << "unsupported filter format at byte " << cb << ": " << (int)dest[cb] << endl;
	    ENDRETURN(false);
	}
    }

    delete [] data;
    data = new float[width*height*4];
    for(int x=0; x < width; x++)
    {
	for(int y=0; y < height; y++)
	{
	    data[(x*height+y)*4+0] = image[(x*height+y)*4+0]/255.0f;
	    data[(x*height+y)*4+1] = image[(x*height+y)*4+1]/255.0f;
	    data[(x*height+y)*4+2] = image[(x*height+y)*4+2]/255.0f;
	    data[(x*height+y)*4+3] = image[(x*height+y)*4+3]/255.0f;
	}
    }
    delete[] image;
    delete[] dest;
    delete[] streamdata;

    ENDRETURN(true);
}

bool Texture::Load(TextureData data, TextureDetails details)
{
    FUNC("Texture::Load(data)");

    glGenTextures(1, &ID);
    glBindTexture(GL_TEXTURE_2D, ID);

    if(details.flags & LINEAR)
    {
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    else
    {
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }

    if(details.flags & XCLAMP)
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    if(details.flags & YCLAMP)
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    GLenum target = GL_TEXTURE_2D;
    GLint format = GL_RGB;
    if(data.Bpp() == 32)
	format = GL_RGBA;

    if(details.flags & FLOAT)
    {
	if(details.flags & LINEAR)
	    glTexImage2D(GL_TEXTURE_2D, 0, format, data.Width(), data.Height(),
			 0, GL_RGBA, GL_FLOAT, data.Data());
	else
	    gluBuild2DMipmaps(target, format, data.Width(), data.Height(),
			      GL_RGBA, GL_FLOAT, data.Data());
    }
    else
    {
	if(HASFLAGS(details.flags, BLACKWHITE | NOALPHA))
	{
	    unsigned char *buf = new unsigned char[data.Width()*data.Height()];

	    for(int x=0; x < data.Width(); x++)
		for(int y=0; y < data.Height(); y++)
		    buf[x*data.Height()+y] = (unsigned char)(data.Data()[(x*data.Height()+y)*4]*255.0f);

	    if(details.flags & LINEAR)
		glTexImage2D(GL_TEXTURE_2D, 0, format, data.Width(), data.Height(),
			     0, GL_LUMINANCE, GL_UNSIGNED_BYTE, buf);
	    else
		gluBuild2DMipmaps(target, format, data.Width(), data.Height(),
				  GL_LUMINANCE, GL_UNSIGNED_BYTE, buf);
	}
	else if(HASFLAGS(details.flags, BLACKWHITE))
	{
	    unsigned char *buf = new unsigned char[data.Width()*data.Height()*3];

	    for(int x=0; x < data.Width(); x++)
		for(int y=0; y < data.Height(); y++)
		{
		    buf[(x*data.Height()+y)*3+0] = (unsigned char)(data.Data()[(x*data.Height()+y)*4+0]*255.0f);
		    buf[(x*data.Height()+y)*3+1] = (unsigned char)(data.Data()[(x*data.Height()+y)*4+3]*255.0f);
		}

	    if(details.flags & LINEAR)
		glTexImage2D(GL_TEXTURE_2D, 0, format, data.Width(), data.Height(),
			     0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, buf);
	    else
		gluBuild2DMipmaps(target, format, data.Width(), data.Height(),
				  GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, buf);
	}
	else if(HASFLAGS(details.flags, NOALPHA))
	{
	    unsigned char *buf = new unsigned char[data.Width()*data.Height()*3];

	    for(int x=0; x < data.Width(); x++)
		for(int y=0; y < data.Height(); y++)
		{
		    buf[(x*data.Height()+y)*3+0] = (unsigned char)(data.Data()[(x*data.Height()+y)*4+0]*255.0f);
		    buf[(x*data.Height()+y)*3+1] = (unsigned char)(data.Data()[(x*data.Height()+y)*4+1]*255.0f);
		    buf[(x*data.Height()+y)*3+2] = (unsigned char)(data.Data()[(x*data.Height()+y)*4+2]*255.0f);
		}

	    if(details.flags & LINEAR)
		glTexImage2D(GL_TEXTURE_2D, 0, format, data.Width(), data.Height(),
			     0, GL_RGB, GL_UNSIGNED_BYTE, buf);
	    else
		gluBuild2DMipmaps(target, format, data.Width(), data.Height(),
				  GL_RGB, GL_UNSIGNED_BYTE, buf);
	}
	else
	{
	    unsigned char *buf = new unsigned char[data.Width()*data.Height()*4];

	    for(int x=0; x < data.Width(); x++)
		for(int y=0; y < data.Height(); y++)
		{
		    buf[(x*data.Height()+y)*4+0] = (unsigned char)(data.Data()[(x*data.Height()+y)*4+0]*255.0f);
		    buf[(x*data.Height()+y)*4+1] = (unsigned char)(data.Data()[(x*data.Height()+y)*4+1]*255.0f);
		    buf[(x*data.Height()+y)*4+2] = (unsigned char)(data.Data()[(x*data.Height()+y)*4+2]*255.0f);
		    buf[(x*data.Height()+y)*4+3] = (unsigned char)(data.Data()[(x*data.Height()+y)*4+3]*255.0f);
		}

	    if(details.flags & LINEAR)
		glTexImage2D(GL_TEXTURE_2D, 0, format, data.Width(), data.Height(),
			     0, GL_RGBA, GL_UNSIGNED_BYTE, buf);
	    else
		gluBuild2DMipmaps(target, format, data.Width(), data.Height(),
				  GL_RGBA, GL_UNSIGNED_BYTE, buf);
	}
    }
    ENDRETURN(true);
}
